import logging
import os
import sys
from typing import Dict

import logmuse
from ubiquerg import VersionInHelpParser

from ._version import __version__
from .assess.cli import build_subparser as assess_subparser
from .bbclient.cli import build_subparser as bbclient_subparser
from .bedspace.cli import build_argparser as bedspace_subparser
from .eval.cli import build_subparser as eval_subparser
from .likelihood.cli import build_subparser as likelihood_subparser
from .region2vec.cli import build_subparser as region2vec_subparser
from .scembed.argparser import build_argparser as scembed_subparser
from .tokenization.cli import build_subparser as tokenization_subparser
from .universe.cli import build_mode_parser as universe_subparser


def print_inspect_beds(bb_cache_folder) -> None:
    """
    Print the bed files in the cache folder.

    :param bb_cache_folder: Cache folder path
    """
    from rich.console import Console
    from rich.table import Table

    from .bbclient import BBClient

    _LOGGER.info(f"Bedfiles directory:")
    bbc = BBClient(cache_folder=bb_cache_folder)
    result = bbc.list_beds()

    console = Console()

    # Create a Table
    table = Table(title="Cached Bedfiles")

    # Add columns
    table.add_column("ID", justify="center", style="cyan", no_wrap=True)
    table.add_column("Path", style="magenta")

    # Add rows from the dictionary
    for id, path in result.items():
        table.add_row(str(id), path)

    # Print the table
    console.print(table)

    console.print(f"Number of bed files: {len(result)}")


def print_inspect_bedsets(bb_cache_folder) -> None:
    """
    Print the bed sets in the cache folder.

    :param bb_cache_folder: Cache folder path
    """
    from rich.console import Console
    from rich.table import Table

    from .bbclient import BBClient

    _LOGGER.info(f"Bedsets directory:")
    bbc = BBClient(cache_folder=bb_cache_folder)
    result = bbc.list_bedsets()

    console = Console()

    # Create a Table
    table = Table(title="Cached Bedsets")

    # Add columns
    table.add_column("ID", justify="center", style="cyan", no_wrap=True)
    table.add_column("Path", style="magenta")

    # Add rows from the dictionary
    for id, path in result.items():
        table.add_row(str(id), path)

    # Print the table
    console.print(table)

    console.print(f"Number of bed sets: {len(result)}")


def build_argparser():
    """
    Builds argument parser.

    :return argparse.ArgumentParser: Argument parser
    """

    banner = "%(prog)s - Genomic Interval toolkit"
    additional_description = "\nhttps://geniml.databio.org"

    parser = VersionInHelpParser(
        prog="geniml",
        version=f"{__version__}",
        description=banner,
        epilog=additional_description,
    )

    # Individual subcommands
    msg_by_cmd = {
        "assess-universe": "Assess a universe",
        "bbclient": "Client for the BEDbase server",
        "bedspace": "Coembed regionsets (bed files) and labels",
        "build-universe": "Build a consensus peak set using one of provided model",
        "eval": "Evaluate a set of region embeddings",
        "lh": "Make likelihood model",
        "region2vec": "Train a region2vec model",
        "scembed": "Embed single-cell data as region vectors",
        "tokenize": "Tokenize BED files",
    }

    sp = parser.add_subparsers(dest="command")
    subparsers: Dict[str, VersionInHelpParser] = {}
    for k, v in msg_by_cmd.items():
        subparsers[k] = sp.add_parser(k, description=v, help=v)

    # build up subparsers for modules
    subparsers["assess-universe"] = assess_subparser(subparsers["assess-universe"])
    subparsers["bbclient"] = bbclient_subparser(subparsers["bbclient"])
    subparsers["bedspace"] = bedspace_subparser(subparsers["bedspace"])
    subparsers["build-universe"] = universe_subparser(subparsers["build-universe"])
    subparsers["eval"] = eval_subparser(subparsers["eval"])
    subparsers["lh"] = likelihood_subparser(subparsers["lh"])
    subparsers["region2vec"] = region2vec_subparser(subparsers["region2vec"])
    subparsers["scembed"] = scembed_subparser(subparsers["scembed"])
    subparsers["tokenize"] = tokenization_subparser(subparsers["tokenize"])

    return parser


def main(test_args=None):
    parser = logmuse.add_logging_options(build_argparser())
    args, _ = parser.parse_known_args()
    if test_args:
        args.__dict__.update(test_args)

    global _LOGGER
    _LOGGER = logmuse.logger_via_cli(args, make_root=True)

    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    _LOGGER.info(f"Command: {args.command}")

    if args.command == "assess-universe":
        from .assess.assess import run_all_assessment_methods

        run_all_assessment_methods(
            args.raw_data_folder,
            args.file_list,
            args.universe,
            args.no_workers,
            args.folder_out,
            args.pref,
            args.save_each,
            args.overlap,
            args.distance,
            args.distance_flexible,
            args.distance_universe_to_file,
            args.distance_flexible_universe_to_file,
        )

    if args.command == "lh":
        # _LOGGER.info(f"Subcommand: {args.subcommand}")
        # if args.subcommand == "build_model":
        from .likelihood.build_model import main

        main(
            args.model_file,
            args.coverage_folder,
            args.coverage_prefix,
            args.file_no,
            args.force,
        )

    if args.command == "bbclient":
        if args.subcommand in [
            "cache-bed",
            "cache-tokens",
            "cache-bedset",
            "seek",
            "inspect-bedfiles",
            "inspect-bedsets",
            "rm",
        ]:
            _LOGGER.info(f"Subcommand: {args.subcommand}")
            from .bbclient import BBClient

            bbc = BBClient(cache_folder=args.cache_folder)

        else:
            # if no subcommand, print help format of bbclient subparser
            # from https://stackoverflow.com/a/20096044/23054783
            import argparse

            subparsers_actions = [
                action
                for action in parser._actions
                if isinstance(action, argparse._SubParsersAction)
            ]
            # there will probably only be one subparser_action,
            # but better safe than sorry
            for subparsers_action in subparsers_actions:
                # get all subparsers and print help
                for choice, subparser in subparsers_action.choices.items():
                    if choice == "bbclient":
                        print(subparser.format_help())
                        sys.exit(1)
        if args.subcommand == "cache-bed":
            # if input is a BED file path
            if os.path.exists(args.identifier[0]):
                identifier = bbc.add_bed_to_cache(args.identifier[0])
                _LOGGER.info(f"BED file {identifier} has been cached")
            else:
                bbc.load_bed(args.identifier[0])

        if args.subcommand == "cache-tokens":
            bbc.add_bed_tokens_to_cache(args.bed_id[0], args.universe_id[0])

        if args.subcommand == "cache-bedset":
            if os.path.isdir(args.identifier[0]):
                from .io import BedSet

                bedset = BedSet(
                    [
                        os.path.join(args.identifier[0], file_name)
                        for file_name in os.listdir(args.identifier[0])
                    ]
                )
                bbc.add_bedset_to_cache(bedset)
                _LOGGER.info(f"BED set {bedset.compute_bedset_identifier()} has been cached")

            else:
                bbc.load_bedset(args.identifier[0])

        if args.subcommand == "seek":
            handler = logging.StreamHandler(sys.stdout)
            _LOGGER.addHandler(handler)
            _LOGGER.info(bbc.seek(args.identifier[0]))

        if args.subcommand == "inspect-bedfiles":
            print_inspect_beds(args.cache_folder)

        if args.subcommand == "inspect-bedsets":
            print_inspect_bedsets(args.cache_folder)

        if args.subcommand == "rm":
            file_path = bbc.seek(args.identifier[0])
            bbc._remove(file_path)
            _LOGGER.info(f"{file_path} is removed")

    if args.command == "build-universe":
        _LOGGER.info(f"Subcommand: {args.subcommand}")
        if args.subcommand == "hmm":
            from .universe.hmm_universe import hmm_universe

            hmm_universe(
                coverage_folder=args.coverage_folder,
                out_file=args.output_file,
                prefix=args.coverage_prefix,
                normalize=args.not_normalize,
                save_max_cove=args.save_max_cove,
            )

        if args.subcommand == "ml":
            from .universe.ml_universe import ml_universe

            ml_universe(
                model_file=args.model_file,
                cove_folder=args.coverage_folder,
                cove_prefix=args.coverage_prefix,
                file_out=args.output_file,
            )

        if args.subcommand == "cc":
            from .universe.cc_universe import cc_universe

            cc_universe(
                cove=args.coverage_folder,
                cove_prefix=args.coverage_prefix,
                file_out=args.output_file,
                merge=args.merge,
                filter_size=args.filter_size,
                cutoff=args.cutoff,
            )
        if args.subcommand == "ccf":
            from .universe.ccf_universe import ccf_universe

            ccf_universe(
                cove=args.coverage_folder,
                cove_prefix=args.coverage_prefix,
                file_out=args.output_file,
            )

    if args.command == "scembed":
        _LOGGER.info("Running scembed")
        pass
        # scembed_main(test_args)

    if args.command == "region2vec":
        from .region2vec import region2vec

        region2vec(
            token_folder=args.token_folder,
            save_dir=args.save_dir,
            num_shufflings=args.num_shuffle,
            num_processes=args.nworkers,
            embedding_dim=args.embed_dim,
            context_win_size=args.context_len,
            save_freq=args.save_freq,
            resume_path=args.resume,
            train_alg=args.train_alg,
            min_count=args.min_count,
            neg_samples=args.neg_samples,
            init_lr=args.init_lr,
            min_lr=args.min_lr,
            lr_scheduler=args.lr_mode,
            milestones=args.milestones,
            seed=args.seed,
        )
    if args.command == "tokenize":
        from .tokenization import hard_tokenization

        hard_tokenization(
            src_folder=args.data_folder,
            dst_folder=args.token_folder,
            universe_file=args.universe,
            fraction=args.fraction,
            num_workers=args.nworkers,
            bedtools_path=args.bedtools_path,
        )
    if args.command == "eval":
        if args.subcommand == "gdst":
            from geniml.eval.gdst import get_gdst_score

            gdst_score = get_gdst_score(
                args.model_path, args.embed_type, args.num_samples, args.seed
            )
            print(gdst_score)
        if args.subcommand == "npt":
            from geniml.eval.npt import get_npt_score

            npt_score = get_npt_score(
                args.model_path,
                args.embed_type,
                args.K,
                args.num_samples,
                args.seed,
                args.K,
                num_workers=args.num_workers,
            )
            print(npt_score["SNPR"][0])
        if args.subcommand == "ctt":
            from geniml.eval.ctt import get_ctt_score

            ctt_score = get_ctt_score(
                args.model_path,
                args.embed_type,
                args.seed,
                args.num_samples,
                args.num_workers,
            )

            print(ctt_score)
        if args.subcommand == "rct":
            from geniml.eval.rct import get_rct_score

            rct_score = get_rct_score(
                args.model_path,
                args.embed_type,
                args.bin_path,
                args.out_dim,
                args.cv_num,
                args.seed,
                args.num_workers,
            )
            print(rct_score)
        if args.subcommand == "bin-gen":
            import glob
            import pickle

            from geniml.eval.utils import get_bin_embeddings

            if os.path.exists(args.file_name):
                print(f"{args.file_name} exists!")
                return
            token_files = glob.glob(os.path.join(args.token_folder, "*"))
            bin_embed = get_bin_embeddings(args.universe, token_files)
            os.makedirs(os.path.dirname(args.file_name), exist_ok=True)
            with open(args.file_name, "wb") as f:
                pickle.dump(bin_embed, f)
            print(f"binary embeddings saved to {args.file_name}")

    return


if __name__ == "__main__":
    main()



import logging
from typing import Union

import numpy as np

from ...const import PKG_NAME
from ...io import RegionSet
from ...region2vec.main import Region2VecExModel
from .abstract import Query2Vec

_LOGGER = logging.getLogger(PKG_NAME)


class BED2Vec(Query2Vec):
    """Embed a query region set into a vector"""

    def __init__(self, model: Union[str, Region2VecExModel]):
        """
        set the Region2VecExModel

        :param model: a Region2VecExModel or a model repository on Hugging Face
        """
        if isinstance(model, str):
            self.model = Region2VecExModel(model)
        elif isinstance(model, Region2VecExModel):
            self.model = model
        else:
            _LOGGER.error(
                "TypeError: Please give a Region2VecExModel or a model repository on Hugging Face"
            )

    def forward(self, query: Union[str, RegionSet]) -> np.ndarray:
        """
        Embed the query region set

        :param query: a RegionSet, or the path to a BED file in disk

        :return: the region set embedding
        """
        # if query is a BED file name, read it as a RegionSet class
        if isinstance(query, str):
            query = RegionSet(query)

        region_embeddings = self.model.encode(query)
        # BED embedding: averaging region embeddings
        return np.mean(region_embeddings, axis=0)



def build_subparser(parser):
    """Builds an argument parser to support the tokenize command line interface."""
    parser.add_argument(
        "--data-folder",
        type=str,
        help="Path to the folder that stores BED files",
    )
    parser.add_argument("--token-folder", type=str, help="Folder that stores tokenized files")
    # parameters for hard tokenization
    parser.add_argument("--universe", type=str, help="Path to a universe file")
    parser.add_argument("--nworkers", type=int, default=10, help="number of workers")
    parser.add_argument(
        "--bedtools-path",
        type=str,
        default="bedtools",
        help="Path to the bedtools binary. Default: bedtools. If bedtools does not exists, an exception will be raised",
    )
    parser.add_argument(
        "--fraction",
        type=float,
        default=1.0e-9,
        help="A parameter for bedtools.intersect",
    )

    return parser



# Project configuration, particularly for logging.

import logmuse

from ._version import __version__
from .bedshift import Bedshift

__classes__ = ["Bedshift"]
__all__ = __classes__ + []

logmuse.init_logger("bedshift")



from typing import Dict, List, Union

import numpy as np

from ..backends import BiVectorBackend
from ..query2vec import Text2Vec
from .abstract import BEDSearchInterface


class BiVectorSearchInterface(BEDSearchInterface):
    """Search interface for ML free bi-vectors searching backend"""

    def __init__(self, backend: BiVectorBackend, query2vec: Union[str, Text2Vec]) -> None:
        """
        Initiate the search interface

        :param backend: the backend where vectors are stored
        :param query2vec: a Text2Vec, for details, see docstrings in geniml.search.query2vec.text2vec
        """
        if isinstance(query2vec, str):
            self.query2vec = Text2Vec(query2vec, v2v=None)
        else:
            self.query2vec = query2vec
        self.backend = backend

    def query_search(
        self,
        query: Union[str, np.ndarray],
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
        p: float = 1.0,
        q: float = 1.0,
        distance: bool = False,
        rank: bool = False,
    ) -> List[Dict]:
        """

        :param query: the natural language query string, or a vector in the embedding space of region sets

        for rest of the parameters, check the docstring of QdrantBackend.search() or HNSWBackend.search()
        """
        if isinstance(query, np.ndarray):
            search_vec = query
        else:
            search_vec = self.query2vec.forward(query)

        return self.backend.search(
            query=search_vec,
            limit=limit,
            with_payload=with_payload,
            with_vectors=with_vectors,
            offset=offset,
            p=p,
            q=q,
            distance=distance,
            rank=rank,
        )



import os
import pickle
from typing import List, Tuple, Union

os.environ["OPENBLAS_NUM_THREADS"] = "1"
import multiprocessing as mp
from multiprocessing.queues import Queue

import numpy as np
from sklearn.linear_model import LinearRegression

from .const import GENOME_DIST_SCALAR
from .utils import cosine_distance, genome_distance, load_genomic_embeddings


def sample_from_vocab(vocab: List[str], num_samples: int, seed: int = 42) -> List[str]:
    """Samples regions from vocab.

    Samples regions proportionally from each chromosome.

    Args:
        vocab (list[str]): A list of regions.
        num_samples (int): Number of regions to sample.
        seed (int, optional): Random seed. Defaults to 42.

    Returns:
        list[str]: A list of sampled regions.
    """
    chr_probs = {}
    region_dict = {}
    num_vocab = len(vocab)
    # build stat from vocab
    for region in vocab:
        chr_str, position = region.split(":")
        chr_str = chr_str.strip()
        start, end = position.split("-")
        start = int(start.strip())
        end = int(end.strip())
        chr_probs[chr_str] = chr_probs.get(chr_str, 0) + 1
        if chr_str in region_dict:
            region_dict[chr_str].append((start, end))
        else:
            region_dict[chr_str] = [(start, end)]
    total = sum([chr_probs[k] for k in chr_probs])
    chr_probs = [(k, chr_probs[k] / total) for k in chr_probs]

    count = 0

    chr_names = [t[0] for t in chr_probs]
    chr_probs = [t[1] for t in chr_probs]
    sampled_regions = []
    np.random.seed(seed)
    while count < num_samples:
        sel_chr = np.random.choice(chr_names, p=chr_probs)
        regions = region_dict[sel_chr]
        if len(regions) < 2:
            continue
        sel_indexes = np.random.choice(len(regions), 2, replace=False)
        r1, r2 = regions[sel_indexes[0]], regions[sel_indexes[1]]
        gdist = genome_distance(r1, r2)
        sampled_regions.append(
            (
                f"{sel_chr}:{r1[0]}-{r1[1]}",
                f"{sel_chr}:{r2[0]}-{r2[1]}",
                gdist,
            )
        )
        count += 1
    return sampled_regions


def get_gdst_score(
    path: str,
    embed_type: str,
    num_samples: int = 10000,
    seed: int = 42,
    queue: Queue = None,
    worker_id: int = None,
) -> Union[float, Tuple[int, str, float]]:
    """Runs the GDST on a model.

    Args:
        path (str): The path to a model.
        embed_type (str): The model type: "region2vec" or "base".
        num_samples (int, optional): Number of embeddings used for evaluation.
            Defaults to 10000.
        seed (int, optional): Random seed. Defaults to 42.
        queue (Queue, optional): The multiprocessing
            queue used to store results. Defaults to None.
        worker_id (int, optional): Worker id. Defaults to None.

    Returns:
        Union[float, tuple[int, str, float]]: A GDST score when used in a single
            process; or, a tuple of (worker id, model path, GDST score) when
            used in multiple processes.
    """
    embed_rep, vocab = load_genomic_embeddings(path, embed_type)
    regions = sample_from_vocab(vocab, num_samples, seed)
    region2idx = {r: i for i, r in enumerate(vocab)}
    gdist_arr = [r[2] / GENOME_DIST_SCALAR for r in regions]
    edist_arr = np.array(
        [
            cosine_distance(embed_rep[region2idx[t[0]]], embed_rep[region2idx[t[1]]])
            for t in regions
        ]
    )
    gd_arr = list(zip(gdist_arr, edist_arr))
    X = np.array([[g[0]] for g in gd_arr])
    y = np.array([g[1] for g in gd_arr])
    reg = LinearRegression().fit(X, y)
    slope = reg.coef_[0]
    if queue:
        queue.put((worker_id, path, slope))
        return worker_id, path, slope
    else:
        return slope


def writer_multiprocessing(save_path: str, num: int, q: Queue) -> List[Tuple[str, float]]:
    """Writes results from multiple processes to a list.

    Args:
        save_path (str): The path to the saved results.
        num (int): The number of results.
        q (Queue): A multiprocessing queue.

    Returns:
        list[tuple[str, float]]: A list of  (model path, GDST score) tuples.
    """
    results = ["" for i in range(num)]
    while True:
        m = q.get()
        if m == "kill":
            break
        index = m[0]
        results[index] = (m[1], m[2])
        if save_path:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            with open(save_path, "wb") as f:
                pickle.dump(results, f)
    return results


def get_gdst_score_batch(
    batch: List[Tuple[str, str]],
    num_samples: int = 10000,
    seed: int = 42,
    save_path: str = None,
    num_workers: int = 1,
) -> List[Tuple[str, float]]:
    """Runs the GDST on a batch of models.

    Args:
        batch (list[tuple[str, str]]): A list of (model path, model type) tuples.
        num_samples (int, optional): Number of embeddings used for evaluation.
            Defaults to 10000.
        seed (int, optional): Random seed. Defaults to 42.
        save_path (str, optional): Save the results to save_path. Defaults to
            None.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 1.

    Returns:
        list[tuple[str, float]]: A list of (model path, GDST score) tuples.
    """
    if num_workers <= 1:
        gds_arr = []
        for path, embed_type in batch:
            gds = get_gdst_score(path, embed_type, num_samples, seed)
            print(path, gds)
            gds_arr.append((path, gds))
        if save_path:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            with open(save_path, "wb") as f:
                pickle.dump(gds_arr, f)
    else:
        manager = mp.Manager()
        queue = manager.Queue()
        with mp.Pool(processes=num_workers) as pool:
            writer = pool.apply_async(writer_multiprocessing, (save_path, len(batch), queue))
            all_processes = []
            for i, (path, embed_type) in enumerate(batch):
                process = pool.apply_async(
                    get_gdst_score,
                    (path, embed_type, num_samples, seed, queue, i),
                )
                all_processes.append(process)

            for process in all_processes:
                process.get()
            queue.put("kill")
            gds_arr = writer.get()
    return gds_arr


def gdst_eval(
    batch: List[Tuple[str, str]],
    num_runs: int = 20,
    num_samples: int = 1000,
    save_folder: str = None,
    num_workers: int = 10,
) -> List[Tuple[str, List[float]]]:
    """Runs the GDST on a batch of models for multiple times.

    Args:
        batch (list[tuple[str, str]]): A list of (model path, model type) tuples.
        num_runs (int, optional): Number of runs. Defaults to 20.
        num_samples (int, optional): Number of embeddings used for evaluation.
            Defaults to 1000.
        save_folder (str, optional): Folder to save the results from each run.
            Defaults to None.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Returns:
        list[tuple[str, list[float]]]: A list of (model path, GDST scores from
            num_runs) tuples.
    """
    results_seeds = []
    for seed in range(num_runs):
        print(f"----------------Run {seed}----------------")
        save_path = os.path.join(save_folder, f"gdst_eval_seed{seed}") if save_folder else None
        result_list = get_gdst_score_batch(batch, num_samples, seed, save_path, num_workers)
        results_seeds.append(result_list)

    gds_res = [[] for i in range(len(batch))]
    for results in results_seeds:
        for i, res in enumerate(results):
            gds_res[i].append(res[1])
            assert res[0] == batch[i][0], "key == batch[i][0]"

    mean_gds = [np.array(r).mean() for r in gds_res]
    std_gds = [np.array(r).std() for r in gds_res]
    models = [t[0] for t in batch]
    for i in range(len(mean_gds)):
        print(f"{batch[i][0]}\n GDST score (std): {mean_gds[i]:.4f} ({std_gds[i]:.4f}) \n")
    gds_arr = [(batch[i][0], gds_res[i]) for i in range(len(batch))]
    return gds_arr



import logging
from typing import Union

import numpy as np

# from langchain_huggingface.embeddings import HuggingFaceEmbeddings
from fastembed import TextEmbedding

from ...const import PKG_NAME
from ...text2bednn import Vec2VecFNN
from .abstract import Query2Vec

_LOGGER = logging.getLogger(PKG_NAME)


class Text2Vec(Query2Vec):
    """Map a query string into a vector into the embedding space of region sets"""

    def __init__(self, hf_repo: str, v2v: Union[str, Vec2VecFNN, None]):
        """
        :param text_embedder: a model repository on Hugging Face
        :param v2v: a Vec2VecFNN (see geniml/text2bednn/text2bednn.py) or a model repository on Hugging Face
        """
        # Set model that embed natural language
        self.text_embedder = TextEmbedding(model_name=hf_repo)
        # Set model that maps natural language embeddings into the embedding space of region sets
        if isinstance(v2v, Vec2VecFNN):
            self.v2v = v2v
        elif isinstance(v2v, str):
            # for bivec search (ML free)
            self.v2v = Vec2VecFNN(v2v)
        else:
            self.v2v = None

    def forward(self, query: str) -> np.ndarray:
        """
        Embed the query natural language string

        :param query: a natural language string

        :return: the embedding vector of query
        """
        # embed query string
        query_embedding = list(self.text_embedder.embed(query))[0]
        if self.v2v is None:
            return query_embedding
        else:
            # map the query string embedding into the embedding space of region sets
            return self.v2v.encode(query_embedding)



def build_subparser(parser):
    """
    Builds argument parser.

    :return argparse.ArgumentParser: Argument parser
    """
    parser.add_argument(
        "--overlap",
        help="if calculate base-level overlap score",
        action="store_true",
    )
    parser.add_argument(
        "--distance",
        help="if calculate distance from region in query to nearest region in the universe",
        action="store_true",
    )
    parser.add_argument(
        "--distance-universe-to-file",
        help="if calculate distance from region in the universe to nearest region query",
        action="store_true",
    )
    parser.add_argument(
        "--distance-flexible",
        help="if calculate distance from region in query to nearest region in the universe taking into account "
        "universe flexibility ",
        action="store_true",
    )
    parser.add_argument(
        "--distance-flexible-universe-to-file",
        help="if calculate distance from region in the universe to nearest region in query taking into account "
        "universe flexibility ",
        action="store_true",
    )
    parser.add_argument(
        "--raw-data-folder",
        help="folder with raw data",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--file-list",
        help="list of files that need to be assessed",
        type=str,
        required=True,
    )
    parser.add_argument("--universe", help="universe file", type=str, required=True)
    parser.add_argument(
        "--no-workers",
        help="number of core that should be used",
        default=4,
        type=int,
    )
    parser.add_argument(
        "--save-to-file",
        help="if save statistics for each BED file to a file",
        action="store_true",
    )
    parser.add_argument("--folder-out", help="folder to which save the statistic", type=str)
    parser.add_argument("--pref", help="statistic file prefix", type=str)

    parser.add_argument(
        "--save-each",
        help="if save distance for each peak in each file ",
        action="store_true",
    )

    return parser



import os

os.environ["OPENBLAS_NUM_THREADS"] = "1"

import multiprocessing as mp
import pickle
from multiprocessing.queues import Queue
from typing import Dict, List, Tuple, Union

import numpy as np

from .utils import genome_distance, load_genomic_embeddings


def get_topk_embed(
    i: int, K: int, embed: np.ndarray, dist: str = "cosine"
) -> Tuple[np.ndarray, np.ndarray]:
    """Gets the nearest K embedding indexes to the i-th embedding.

    Args:
        i (int): The index for the query embedding.
        K (int): The number of nearest embeddings to select.
        embed (np.ndarray): An array of embedding vectors
        dist (str, optional): The distance function used. Defaults to "cosine".

    Returns:
        tuple[np.ndarray, np.ndarray]: K indexes of nearest embeddings and the
            corresponding similarities.
    """
    num = len(embed)
    if dist == "cosine":
        nom = np.dot(embed[i : i + 1], embed.T)
        denom = np.linalg.norm(embed[i : i + 1]) * np.linalg.norm(embed, axis=1)
        sims = (nom / denom)[0]
        indexes = np.argsort(-sims)[1 : K + 1]
        s = sims[indexes]
    elif dist == "euclidean":
        dist = np.linalg.norm(embed[i : i + 1] - embed, axis=1)
        indexes = np.argsort(dist)[1 : K + 1]
        s = -dist[indexes]
    elif dist == "jaccard":
        nom = np.dot(embed[i : i + 1], embed.T)
        denom = ((embed[i : i + 1] + embed) > 0.0).sum(axis=1)
        sims = (nom / denom)[0]
        indexes = np.argsort(-sims)[1 : K + 1]
        s = sims[indexes]
    return indexes, s


def find_kneighbors(region_array: List[Tuple[str, int, int]], index: int, k: int) -> List[int]:
    """Finds the indexes of the K nearest regions of a query region on genome.

    region_array must be sorted, and all regions are on the same chromosome.

    Args:
        region_array (list[tuple[str, int, int]]): A list of (chromosome, start
            position, end position) tuples.
        index (int): The index of the query region.
        k (int): Specifies the number of nearest neighbors of the query region.

    Returns:
        list[int]: A list of indexes of the K nearest neighbors in
            region_array.
    """
    if len(region_array) < k:
        k = len(region_array)
    qregion = region_array[index]
    left_idx = max(index - k, 0)
    right_idx = min(index + k, len(region_array) - 1)
    rdist_arr = []
    for idx in range(left_idx, right_idx + 1):
        rdist_arr.append(genome_distance(qregion, region_array[idx]))
    rdist_arr = np.array(rdist_arr)
    Kneighbors_idx = np.argsort(rdist_arr)[1 : k + 1]
    Kneighbors_idx = Kneighbors_idx + left_idx
    return Kneighbors_idx


def calculate_overlap_bins(
    local_idx: int,
    K: int,
    chromo: str,
    region_array: List[Tuple[str, int, int]],
    region2index: dict[str, int],
    embed_rep: np.ndarray,
    res: int = 10,
    dist: str = "cosine",
    same_chromo: bool = True,
) -> np.ndarray:
    """Calculates the overlap ratios for a region.

    Calculates the overlap ratios for a region between its K-nearest neighbor
    set obtained using genome distance and its K-nearest neighbor set obtained
    using embedding distance. If res < K, then calculates ratios for size
    res*1, res*2, ..., min(res*n, K).

    Args:
        local_idx (int): The local index of a region on its chromosome.
        K (int): Specifies the number of nearest neighbors.
        chromo (str): Chromosome.
        region_array (list[tuple[str, int, int]]): A list of (chromosome, start
            position, end position) tuples.
        region2index (dict[str, int]): A dictionary of (region, index).
        embed_rep (np.ndarray): An array of embedding vectors.
        res (int, optional): Resolution. Size of neighborhood set. Defaults to
            10.
        dist (str, optional): Distance function. Defaults to "cosine".
        same_chromo (bool, optional): Whether to find nearest neighbors on the
            same chromosome in the embedding space. Defaults to True.

    Returns:
        np.ndarray: An array of overlap ratios.
    """
    Kindices = find_kneighbors(region_array, local_idx, K)
    if len(Kindices) == 0:
        return 0
    str_kregions = [
        f"{chromo}:{region_array[k][0]}-{region_array[k][1]}" for k in Kindices
    ]  # sorted in ascending order
    _Krdist_global_indices = np.array([region2index[r] for r in str_kregions])

    if same_chromo:
        chr_regions = [
            f"{chromo}:{region_array[k][0]}-{region_array[k][1]}" for k in range(len(region_array))
        ]
        chr_global_indices = np.array([region2index[r] for r in chr_regions])
        chr_embeds = embed_rep[chr_global_indices]
        _Kedist_local_indices, _ = get_topk_embed(local_idx, K, chr_embeds, dist)
        _Kedist_global_indices = np.array([chr_global_indices[i] for i in _Kedist_local_indices])
    else:
        idx = region2index[f"{chromo}:{region_array[local_idx][0]}-{region_array[local_idx][1]}"]
        _Kedist_global_indices, _ = get_topk_embed(idx, K, embed_rep, dist)

    bin_overlaps = []
    prev = 0
    assert res < K + 1, "resolution < K + 1"
    for i in range(res, K + 1, res):
        set1 = set(_Krdist_global_indices[prev:i])
        set2 = set(_Kedist_global_indices[prev:i])

        overlap = len(set1.intersection(set2)) / len(set1)
        bin_overlaps.append(overlap)

    return np.array(bin_overlaps)


def cal_snpr(ratio_embed: np.ndarray, ratio_random: np.ndarray) -> np.ndarray:
    """Calculates SNPR values.

    :param ratio_embed: Overlap ratios for query embeddings.
    :param ratio_random: Overlap ratios for random embeddings.

    :return: SNPR values.
    """
    res = np.log10((ratio_embed + 1.0e-10) / (ratio_random + 1.0e-10))
    res = np.maximum(res, 0)
    return res


var_dict = {}


def worker_func(
    i: int,
    K: int,
    chromo: str,
    region_array: List[Tuple[str, int, int]],
    embed_type: str,
    resolution: int,
    dist: str,
) -> np.ndarray:
    """Wrapper for calculate_overlap_bins

    Args:
        i (int): The local index of a region on its chromosome.
        K (int): Specifies the number of nearest neighbors.
        chromo (str): Chromosome.
        region_array (list[tuple[str, int, int]]): A list of (chromosome, start
            position, end position) tuples.
        embed_type (str): Embedding type, "region2vec" or "base".
        resolution (int): Resolution.
        dist (str): Distance function.

    Returns:
        np.ndarray: An array of overlap ratios.
    """
    if embed_type == "embed":
        embeds = var_dict["embed_rep"]
    elif embed_type == "random":
        embeds = var_dict["ref_embed"]
    nprs = calculate_overlap_bins(
        i,
        K,
        chromo,
        region_array,
        var_dict["region2vec_index"],
        embeds,
        resolution,
        dist,
    )
    return nprs


def init_worker(
    embed_rep: np.ndarray, ref_embed: np.ndarray, region2index: Dict[str, int]
) -> None:
    """Initializes data used by workers.

    Args:
        embed_rep (np.ndarray): Query embeddings.
        ref_embed (np.ndarray): Random embeddings.
        region2index (dict[str, int]): A region to index dictionary.
    """
    var_dict["embed_rep"] = embed_rep
    var_dict["ref_embed"] = ref_embed
    var_dict["region2vec_index"] = region2index


def get_npt_score(
    model_path: str,
    embed_type: str,
    K: int,
    num_samples: int = 100,
    seed: int = 0,
    resolution: int = 10,
    dist: str = "cosine",
    num_workers: int = 10,
) -> Dict[str, Union[int, np.ndarray, str]]:
    """Runs the NPT on a model.

    If num_samples > 0, then randomly sample num_samples regions proportional
    from each chromosome. If num_samples == 0, all regions are used in the
    test. If K > resolution, then returns an array of NPT scores; otherwise,
    returns one NPT score.

    Args:
        model_path (str): The path to a model.
        embed_type (str):  The model type: "region2vec" or "base".
        K (int): Specifies the number of nearest neighbors.
        num_samples (int, optional): Number of embeddings used for evaluation.
            Defaults to 100.
        seed (int, optional): Random seed. Defaults to 0.
        resolution (int, optional): Resolution of a neighborhood set. Defaults
            to 10.
        dist (str, optional): Distance function. Defaults to "cosine".
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Returns:
        dict[str, Union[int, np.ndarray, str]]: NPT results in a dictionary
            "K": K,
            "Avg_qNPR": Average NPR ratios for query embeddings,
            "Avg_rNPR": Average NPR ratios for random embeddings,,
            "SNPR": SNPR values,
            "Resolution": Resolution,
            "Path": Model path,
    """
    embed_rep, regions_r2v = load_genomic_embeddings(model_path, embed_type)

    region2index = {r: i for i, r in enumerate(regions_r2v)}
    # Group regions by chromosomes
    chromo_regions = {}
    for v in regions_r2v:
        chromo, region = v.split(":")  # e.g. chr1:100-1000
        chromo = chromo.strip()  # remove possible spaces
        region = region.strip()  # remove possible spaces
        start, end = region.split("-")
        start = int(start.strip())
        end = int(end.strip())
        if chromo not in chromo_regions:
            chromo_regions[chromo] = [(start, end)]
        else:
            chromo_regions[chromo].append((start, end))

    # sort regions in each chromosome
    chromo_ratios = {}
    for chromo in chromo_regions:
        region_array = chromo_regions[chromo]
        chromo_regions[chromo] = sorted(region_array, key=lambda x: x[0])
        chromo_ratios[chromo] = len(region_array) / len(regions_r2v)

    num_regions, num_dim = embed_rep.shape

    np.random.seed(seed)

    ref_embed = (np.random.rand(num_regions, num_dim) - 0.5) / num_dim

    avg_ratio = 0.0
    avg_ratio_ref = 0.0
    count = 0

    if num_workers > 1:
        with mp.Pool(
            processes=num_workers,
            initializer=init_worker,
            initargs=(embed_rep, ref_embed, region2index),
        ) as pool:
            all_processes = []
            for chromo in chromo_regions:
                region_array = chromo_regions[chromo]
                if num_samples == 0:  # exhaustive
                    indexes = list(range(len(region_array)))
                else:
                    num = min(
                        len(region_array),
                        round(num_samples * chromo_ratios[chromo]),
                    )
                    indexes = np.random.permutation(len(region_array))[0:num]
                for i in indexes:
                    process_embed = pool.apply_async(
                        worker_func,
                        (i, K, chromo, region_array, "embed", resolution, dist),
                    )
                    process_random = pool.apply_async(
                        worker_func,
                        (
                            i,
                            K,
                            chromo,
                            region_array,
                            "random",
                            resolution,
                            dist,
                        ),
                    )
                    all_processes.append((process_embed, process_random))

            for i, (process_embed, process_random) in enumerate(all_processes):
                avg_ratio = (avg_ratio * count + process_embed.get()) / (count + 1)
                avg_ratio_ref = (avg_ratio_ref * count + process_random.get()) / (count + 1)
                count = count + 1
    else:
        for chromo in chromo_regions:
            region_array = chromo_regions[chromo]
            if num_samples == 0:  # exhaustive
                indexes = list(range(len(region_array)))
            else:
                num = min(
                    len(region_array),
                    round(num_samples * chromo_ratios[chromo]),
                )
                indexes = np.random.permutation(len(region_array))[0:num]
            for i in indexes:
                nprs_embed = calculate_overlap_bins(
                    i,
                    K,
                    chromo,
                    region_array,
                    region2index,
                    embed_rep,
                    resolution,
                    dist,
                )
                nprs_random = calculate_overlap_bins(
                    i,
                    K,
                    chromo,
                    region_array,
                    region2index,
                    ref_embed,
                    resolution,
                    dist,
                )
                avg_ratio = (avg_ratio * count + nprs_embed) / (count + 1)
                avg_ratio_ref = (avg_ratio_ref * count + nprs_random) / (count + 1)
                count = count + 1
    snprs = cal_snpr(avg_ratio, avg_ratio_ref)

    ratio_msg = " ".join([f"{r:.6f}" for r in avg_ratio])
    ratio_ref_msg = " ".join([f"{r:.6f}" for r in avg_ratio_ref])
    snprs_msg = " ".join([f"{r:.6f}" for r in snprs])
    result = {
        "K": K,
        "Avg_qNPR": avg_ratio,
        "Avg_rNPR": avg_ratio_ref,
        "SNPR": snprs,
        "Resolution": resolution,
        "Path": model_path,
    }
    return result


def writer_multiprocessing(save_path: str, num: int, q: Queue) -> List[Tuple[str, float]]:
    """Writes results from multiple processes to a list.

    Args:
        save_path (str): The path to the saved results.
        num (int): The number of results.
        q (Queue): A multiprocessing queue.

    Returns:
        list[tuple[str, float]]: A list of (model path, NPT score) tuples.
    """
    results = [[] for i in range(num)]
    while True:
        m = q.get()
        if m == "kill":
            break
        worker_id = m[0]
        results[worker_id] = m[1]
        if save_path:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            with open(save_path, "wb") as f:
                pickle.dump(results, f)
    return results


def get_npt_score_batch(
    batch: List[Tuple[str, str]],
    K: int,
    num_samples: int = 100,
    num_workers: int = 10,
    seed: int = 0,
    resolution: int = 10,
    dist: str = "cosine",
    save_path: str = None,
) -> List[Dict[str, Union[int, np.ndarray, str]]]:
    """Runs the NPT on a batch of models.

    Args:
        batch (list[tuple[str, str]]): A list of (model path, model type) tuples.
        K (int): Specifies the number of nearest neighbors.
        num_samples (int, optional): Number of embeddings used for evaluation.
            Defaults to 100.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.
        seed (int, optional): Random seed. Defaults to 0.
        resolution (int, optional): Resolution of a neighborhood set. Defaults
            to 10.
        dist (str, optional): Distance function. Defaults to "cosine".
        save_path (str, optional): Save the results to save_path. Defaults to
            None.

    Returns:
        list[dict[str, Union[int, np.ndarray, str]]]: A list of dictionaries of
            NPT results.
    """
    result_list = []
    for index, (path, embed_type) in enumerate(batch):
        result = get_npt_score(
            path,
            embed_type,
            K,
            num_samples,
            seed,
            resolution,
            dist,
            num_workers,
        )
        result_list.append(result)
    if save_path:
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        with open(save_path, "wb") as f:
            pickle.dump(result_list, f)
    return result_list


def npt_eval(
    batch: List[Tuple[str, str]],
    K: int,
    num_samples: int = 100,
    num_workers: int = 10,
    num_runs: int = 20,
    resolution: int = 10,
    dist: str = "cosine",
    save_folder: str = None,
) -> List[Tuple[str, np.ndarray, int]]:
    """Runs the NPT on a batch of models for multiple times.

    Args:
        batch (list[tuple[str, str]]): A list of (model path, model type) tuples.
        K (int): Specifies the number of nearest neighbors.
        num_samples (int, optional): Number of embeddings used for evaluation.
            Defaults to 100.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.
        num_runs (int, optional): Number of runs. Defaults to 20.
        resolution (int, optional): Resolution of a neighborhood set.
            Defaults to 10.
        dist (str, optional): Distance function. Defaults to "cosine".
        save_folder (str, optional): Folder to save the results from each run.
            Defaults to None.

    Returns:
        list[tuple[str, np.ndarray, int]]: A list of (model path, snprs from
            num_runs, resolution) tuples.
    """
    results_seeds = []
    assert resolution <= K, "resolution <= K"
    for seed in range(num_runs):
        print(f"----------------Run {seed}----------------")
        save_path = os.path.join(save_folder, f"npt_eval_seed{seed}") if save_folder else None
        result_list = get_npt_score_batch(
            batch,
            K,
            num_samples=num_samples,
            num_workers=num_workers,
            seed=seed,
            resolution=resolution,
            dist=dist,
            save_path=save_path,
        )
        results_seeds.append(result_list)
    snpr_results = [[] for i in range(len(batch))]
    paths = ["" for i in range(len(batch))]
    for results in results_seeds:
        for i, result in enumerate(results):
            key = result["Path"]
            snpr_results[i].append(result["SNPR"])
            paths[i] = key
    snpr_results = [np.array(v) for v in snpr_results]
    for i in range(len(batch)):
        snpr_arr = snpr_results[i]
        avg_snprs = snpr_arr.mean(axis=0)
        std_snprs = snpr_arr.std(axis=0)
        msg = " ".join([f"{m:.4f}({s:.4f})" for m, s in zip(avg_snprs, std_snprs)])
        print(f"{paths[i]}\nSNPRs:{msg}\n")
    snpr_results = [(paths[i], snpr_results[i], resolution) for i in range(len(batch))]
    return snpr_results



from typing import Optional

from ..exceptions import GenimlBaseError


class BackedFileNotAvailableError(GenimlBaseError):
    default_message = "File from url is not available in backed mode."

    def __init__(self, message: Optional[str] = None):
        super().__init__(message or self.default_message)


class BEDFileReadError(GenimlBaseError):
    default_message = "Error reading BED file."

    def __init__(self, message: Optional[str] = None):
        super().__init__(message or self.default_message)



import os
import shlex
import subprocess
import tempfile
from subprocess import PIPE, Popen


def prep_data(folder, file, tmp_file):
    """File sort and merge"""
    fin = os.path.join(folder, file)
    stdin_1 = None
    if ".gz" in file:
        command0 = f"zcat {fin}"
        fin = ""
        p0 = Popen(shlex.split(command0), stdout=PIPE)
        stdin_1 = p0.stdout
    command1 = f"sort -k1,1V -k2,2n {fin}"
    command2 = "bedtools merge"
    p1 = Popen(shlex.split(command1), stdin=stdin_1, stdout=PIPE)
    p2 = Popen(shlex.split(command2), stdin=p1.stdout, stdout=PIPE)
    p1.stdout.close()  # Allow p1 to receive a SIGPIPE if p2 exits.
    output = p2.communicate()[0]
    tmp_file.write(output)
    tmp_file.seek(0)


def check_if_uni_sorted(universe):
    """Check if regions in file are sorted"""
    command0 = f"sort -k1,1V -k2,2n -c {universe}"
    output = subprocess.run(shlex.split(command0))
    if output.returncode:
        raise Exception("Universe not sorted")


def check_if_uni_flexible(universe):
    with open(universe) as u:
        l = u.readline()
        l = l.split("\t")
        if len(l) < 6:
            raise Exception("Universe is not flexible")


def process_line(line):
    """Helper for reading in bed file line"""
    line = line.split("\t")[:3]
    chrom = line[0]
    pos = [int(i) for i in line[1:]]
    start = pos[0]
    return pos, start, chrom


def chrom_cmp_bigger(a, b):
    """Natural check if chromosomes name is bigger"""
    ac = a.replace("chr", "")
    ac = ac.split("_")[0]
    bc = b.replace("chr", "")
    bc = bc.split("_")[0]
    if bc.isnumeric() and ac.isnumeric() and bc != ac:
        if int(bc) < int(ac):
            return True
        else:
            return False
    else:
        if b < a:
            return True
        else:
            return False


def process_db_line(dn, pos_index):
    """Helper for reading in universe bed file line"""
    dn = dn.split("\t")
    region_chrom = dn[0]
    return [int(dn[p]) for p in pos_index], region_chrom



import logging
from typing import Dict, List, Set, Union

import numpy as np
import pandas as pd

try:
    import torch
    from torch.utils.data import DataLoader, TensorDataset
except ImportError:
    raise ImportError(
        "Please install Machine Learning dependencies by running 'pip install geniml[ml]'"
    )

from .const import (
    DEFAULT_BATCH_SIZE,
    DEFAULT_DATALOADER_SHUFFLE,
    DEFAULT_FILE_KEY,
    DEFAULT_GENOME_KEY,
    DEFAULT_SERIES_KEY,
    MODULE_NAME,
)

_LOGGER = logging.getLogger(MODULE_NAME)


def arrays_to_torch_dataloader(
    X: np.ndarray,
    Y: np.ndarray,
    target: np.ndarray,
    batch_size: int = DEFAULT_BATCH_SIZE,
    shuffle: bool = DEFAULT_DATALOADER_SHUFFLE,
) -> DataLoader:
    """
    Based on https://stackoverflow.com/questions/44429199/how-to-load-a-list-of-numpy-arrays-to-pytorch-dataset-loader
    Store np.ndarray of X and Y into a torch.DataLoader

    :param X: embedding vectors of input data (natural language embeddings)
    :param Y: embedding vectors of output data (BED file embeddings)
    :param target: vector of 1 and -1, indicating if each vector pair of (X, Y) are target pairs or not
    :param batch_size: size of small batch
    :param shuffle: shuffle dataset or not
    :return: a Dataset for pytorch training in format of torch.DataLoader
    """
    tensor_X = torch.from_numpy(dtype_check(X))
    tensor_Y = torch.from_numpy(dtype_check(Y))
    tensor_target = torch.from_numpy(dtype_check(target))
    my_dataset = TensorDataset(tensor_X, tensor_Y, tensor_target)  # create your dataset

    return DataLoader(my_dataset, batch_size=batch_size, shuffle=shuffle)


def dtype_check(vecs: np.ndarray) -> np.ndarray:
    """
    Since the default float in np is float64, but in pytorch tensor it's float32,
    to avoid errors, the dtype will be switched

    :param vecs: input np.ndarray

    :return: np.ndarray with dtype of float32
    """
    if not isinstance(vecs.dtype, type(np.dtype("float32"))):
        vecs = vecs.astype(np.float32)

    return vecs


def metadata_dict_from_csv(
    csv_path: str,
    col_names: Set[str],
    file_key: str = DEFAULT_FILE_KEY,
    genomes: Union[Set[str], None] = None,
    genomes_key: Union[str, None] = DEFAULT_GENOME_KEY,
    series_key: Union[str, None] = DEFAULT_SERIES_KEY,
    chunk_size: Union[int, None] = None,
) -> Dict[str, Union[str, Dict[str, Union[str, List[str]]]]]:
    """
    Read selected columns from a metadata csv and return metadata dictionary,
    can filter genomes with given list of genomes and the column name of genome

    :param csv_path: path to the csv file that contain metadata
    :param col_names: set of csv columns that contain informative metadata
    :param file_key: name of column of file names
    :param genomes: set of genomes
    :param genomes_key: name of column of sample genomes
    :param series_key: name of column of series
    :param chunk_size: size of chunk to read when the csv file is large

    :return: a metadata dictionary in this format:
    if series information is in the csv, the dictionary format will be:
    {
        <series>:[
            {
                "name": <file name>
                "metadata": {
                    <csv column name>: <metadata string>,
                    ...
                }
            },
            ...
        ],
        ...
    }

    else, the dictionary format will be:
    {
        <file name>: {
            <csv column name>: <metadata string>,
            ...
        },
        ...
    }
    """

    # dictionary to store data
    output_dict = dict()
    # count number of series, files, and csv chunks
    series_count = 0
    bed_count = 0
    text_count = 0
    empty_count = 0
    read_chunk = True
    # read csv
    for chunk in pd.read_csv(csv_path, chunksize=chunk_size):
        # if chunk size is None
        if isinstance(chunk, str):
            read_chunk = False
            rows_to_ite = pd.read_csv(csv_path)
        else:
            rows_to_ite = chunk

        for index, row in rows_to_ite.iterrows():
            genome_filter = True
            # select genome if list of genomes and genome key are given
            if genomes is not None and genomes_key is not None:
                if row[genomes_key].strip() not in genomes:
                    genome_filter = False

            if genome_filter:
                # collect metadata
                metadata_dict = dict()

                for col in col_names:
                    if isinstance(row[col], str):  #
                        text_count += 1
                        metadata_dict[col] = row[col]

                if len(metadata_dict) == 0:
                    empty_count += 1
                # add the metadata into output dictionary
                else:
                    if series_key is None or not series_key in rows_to_ite.columns:
                        output_dict[row[file_key]] = metadata_dict

                    else:
                        payload = {
                            "name": row[file_key],
                            "metadata": metadata_dict,
                        }
                        try:
                            output_dict[row[series_key]].append(payload)
                        except:
                            output_dict[row[series_key]] = [payload]
                            series_count += 1
                bed_count += 1
        if not read_chunk:
            break

    # output of summary statistics
    if series_key is not None:
        _LOGGER.info(f"Number of series: {series_count}")

    _LOGGER.info(f"Number of files: {bed_count}")
    _LOGGER.info(f"Number of metadata strings: {text_count}")
    _LOGGER.info(f"Number of files with 0 metadata strings: {empty_count}")

    return output_dict



import logging
import math
from typing import Dict, List, Tuple, Union

import numpy as np

from ...const import PKG_NAME
from .abstract import EmSearchBackend

_LOGGER = logging.getLogger(PKG_NAME)


def batch_for_request(
    ids: Union[List[str], List[int]], ranks_scores: List[float], batch_size: int = 100
) -> List[Tuple[List, List]]:  # used ChatGPT
    """

    :param ids: collected ids of BED files matching retrieved metadata tags
    :param ranks_scores: keep track of ranks or scores from metadata tag embedding vector search
    :param batch_size: size of batch, > 100 may crash qdrant server
    :return: batched BED ids and matching text-search scores/ranks
    """
    # Check if the lists are the same length
    if len(ids) != len(ranks_scores):
        raise ValueError("The lists must have the same length.")

    # Create batches
    batches = []
    for i in range(0, len(ids), batch_size):
        batch1 = ids[i : i + batch_size]
        batch2 = ranks_scores[i : i + batch_size]
        batches.append((batch1, batch2))

    return batches


class BiVectorBackend:
    """
    Search backend that connects the embeddings of metadata tags and bed files
    """

    def __init__(
        self,
        metadata_backend: EmSearchBackend,
        bed_backend: EmSearchBackend,
        metadata_payload_matches: str = "matched_files",
    ):
        """
        :param metadata_backend: search backend where embedding vectors of metadata tags are stored
        :param bed_backend: search backend where embedding vectors of BED files are stored
        :param metadata_payload_matches: the key in metadata backend payloads to files matching to that metadata tag
        """
        self.metadata_backend = metadata_backend
        self.bed_backend = bed_backend
        self.metadata_payload_matches = metadata_payload_matches

    def search(
        self,
        query: np.ndarray,
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
        p: float = 1.0,
        q: float = 1.0,
        distance: bool = False,
        rank: bool = False,
    ) -> List[Dict[str, Union[int, float, Dict[str, str], List[float]]]]:
        """
        :param query: query vector (embedding vector of query term)
        :param limit: number of nearest neighbors to search for query vector
        :param with_payload: whether payload is included in the result
        :param with_vectors: whether the stored vector is included in the result
        :param offset: the offset of the search results
        :param p: weights to the score of metadata search, recommend 0 < p <= 1.0
        :param q: weights to the score of BED search, recommend 0 < q <= 1.0
        :param distance: whether the score is distance or similarity
        :param rank: whether the result is ranked based on rank or score
        :return: the search result(a list of dictionaries,
            each dictionary include: storage id, vector payload (optional), vector (optional))
        """

        # the key for the score in result: distance or score (cosine similarity)
        self.score_key = "distance" if distance else "score"

        # metadata search
        metadata_results = self.metadata_backend.search(
            query,
            limit=int(math.log(limit) * 5) if limit > 10 else 5,
            with_payload=True,
            offset=0,
        )

        if isinstance(metadata_results, dict):
            metadata_results = [metadata_results]

        if rank:
            return self._rank_search(metadata_results, limit, with_payload, with_vectors, offset)
        else:
            return self._score_search(
                metadata_results, limit, with_payload, with_vectors, offset, p, q
            )

    def _rank_search(
        self,
        metadata_results: List[Dict],
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> List[Dict[str, Union[int, float, Dict[str, str], List[float]]]]:
        """
        Search based on maximum rank in results of metadata embedding and results of BED embedding

        :param metadata_results: result of metadata search
        :param limit: see docstring of def search
        :param with_payload:
        :param with_vectors:
        :param offset:
        :return: the search result ranked based on maximum rank
        """

        text_rank = []
        ids_to_retrieve = []

        query_bed_ids = set()

        for i, result in enumerate(metadata_results):
            # all bed files matching the retrieved metadata tag
            bed_ids = result["payload"][self.metadata_payload_matches]

            unique_bed_ids = [id_ for id_ in bed_ids if id_ not in query_bed_ids]
            query_bed_ids.update(unique_bed_ids)

            for id_ in unique_bed_ids:
                text_rank.append(i)
                ids_to_retrieve.append(id_)
        bed_results = []
        max_rank = []
        request_batches = batch_for_request(ids_to_retrieve, text_rank, 100)

        for ids, ranks in request_batches:
            query_beds = self.bed_backend.retrieve_info(ids, with_vectors=True)
            if isinstance(query_beds, dict):
                query_beds = [query_beds]

            bed_vecs = [b["vector"] for b in query_beds]

            # search request once
            retrieved_batch = self.bed_backend.search(
                np.array(bed_vecs),
                limit=limit,
                with_payload=with_payload,
                with_vectors=with_vectors,
                offset=0,
            )

            for i, retrieved_beds in enumerate(retrieved_batch):
                # j: rank for each bed vector query search
                for j, retrieval in enumerate(retrieved_beds):
                    bed_results.append(retrieval)
                    # collect maximum rank
                    max_rank.append(max(ranks[i], j))

        return self._top_k(max_rank, bed_results, limit, offset=offset, rank=True)

    def _score_search(
        self,
        metadata_results: List[Dict],
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
        p: float = 1.0,
        q: float = 1.0,
    ) -> List[Dict[str, Union[int, float, Dict[str, str], List[float]]]]:
        """
        Search based on weighted score from results of metadata embedding and results of BED embedding

        :param metadata_results: result of metadata search
        :param limit: see docstring of def search
        :param with_payload:
        :param with_vectors:
        :param offset:
        :param p:
        :param q:
        :return: the search result ranked based on weighted similarity scores
        """
        text_scores = []
        ids_to_retrieve = []

        query_bed_ids = set()
        for i, result in enumerate(metadata_results):
            # all bed files matching the retrieved metadata tag
            text_score = (
                1 - result[self.score_key]
                if self.score_key == "distance"
                else result[self.score_key]
            )
            bed_ids = result["payload"][self.metadata_payload_matches]

            unique_bed_ids = [id_ for id_ in bed_ids if id_ not in query_bed_ids]
            query_bed_ids.update(unique_bed_ids)

            for id_ in unique_bed_ids:
                text_scores.append(text_score)
                ids_to_retrieve.append(id_)

        bed_results = []
        overall_scores = []

        request_batches = batch_for_request(ids_to_retrieve, text_scores, 100)

        for ids, scores in request_batches:
            query_beds = self.bed_backend.retrieve_info(ids, with_vectors=True)
            if isinstance(query_beds, dict):
                query_beds = [query_beds]

            bed_vecs = [b["vector"] for b in query_beds]

            retrieved_batch = self.bed_backend.search(
                np.array(bed_vecs),
                limit=limit,
                with_payload=with_payload,
                with_vectors=with_vectors,
                offset=0,
            )

            for i, retrieved_beds in enumerate(retrieved_batch):
                # j: rank for each bed vector query search
                for retrieval in retrieved_beds:
                    # calculate weighted score
                    bed_score = (
                        1 - retrieval[self.score_key]
                        if self.score_key == "distance"
                        else retrieval[self.score_key]
                    )
                    bed_results.append(retrieval)
                    overall_scores.append((p * scores[i] + q * bed_score) / 2)

        return self._top_k(overall_scores, bed_results, limit=limit, offset=offset, rank=False)

    def _top_k(
        self,
        scales: List[Union[int, float]],
        results: List[Dict[str, Union[int, float, Dict[str, str], List[float]]]],
        limit: int = 10,
        offset: int = 0,
        rank: bool = True,
    ):
        """
        Sort top k result and remove repetition

        :param scales: list of weighted scores or maximum rank
        :param results: retrieval result
        :param limit: number of result to return
        :param offset: the offset of the search results
        :param rank: whether the scale is maximum rank or not
        :return: the top k selected result after rank
        """
        paired_score_results = list(zip(scales, results))

        # sort result
        if not rank:
            paired_score_results.sort(reverse=True, key=lambda x: x[0])
        else:
            paired_score_results.sort(key=lambda x: x[0])

        unique_result = {}
        for scale, result in paired_score_results:
            store_id = result["id"]
            # filter out overlap
            if store_id not in unique_result:
                # add rank or score into the result
                if not rank:
                    if self.score_key == "distance":
                        del result[self.score_key]
                    result["score"] = scale
                else:
                    try:
                        del result["score"]
                    except KeyError:
                        del result["distance"]

                    result["max_rank"] = scale
                unique_result[store_id] = result

        top_k_results = list(unique_result.values())[offset : limit + offset]
        return top_k_results



import lightning as L
import pytest
import torch
from torch.utils.data import DataLoader

from geniml.atacformer.main import Atacformer, AtacformerExModel
from geniml.atacformer.utils import AtacformerMLMCollator, AtacformerMLMDataset
from geniml.tokenization.main import AnnDataTokenizer

# from geniml.training.adapters import MLMAdapter


@pytest.fixture
def universe_file():
    return "tests/data/universe_mlm.bed"


@pytest.fixture
def data():
    return "tests/data/gtok_sample/"


@pytest.mark.skip("Too new to test")
def test_atacformer_dataset():
    # t = AnnDataTokenizer("/Users/nathanleroy/Desktop/screen.bed")
    # path_to_data = "/Users/nathanleroy/Desktop/gtoks"
    # path_to_data = "tests/data/gtok_sample/"

    t = AnnDataTokenizer("tests/data/universe_mlm.bed")
    path_to_data = "tests/data/gtok_sample/"

    dataset = AtacformerMLMDataset(path_to_data, t.mask_token_id(), len(t))

    assert dataset is not None
    assert all([isinstance(x, tuple) for x in dataset])


@pytest.mark.skip("Too new to test")
def test_atacformer_init():
    model = Atacformer(
        10_000,  # vocab_size of 10,000 regions
    )
    assert model is not None

    input = torch.randint(0, 10_000, (32, 128))
    output = model(input)
    assert output.shape == (32, 128, 768)


@pytest.mark.skip("Too new to test")
def test_atacformer_exmodel_init(universe_file: str):
    tokenizer = AnnDataTokenizer(universe_file)
    model = AtacformerExModel(
        tokenizer=tokenizer,
    )

    # these are the defaults
    assert model._model.d_model == 768
    assert model._model.vocab_size == 2436
    assert model._model.nhead == 8
    assert model._model.num_layers == 6


@pytest.mark.skip("Too new to test")
def test_train_atacformer_ex_model(universe_file: str, data: str):
    # make tokenizer and model
    tokenizer = AnnDataTokenizer(universe_file)
    model = AtacformerExModel(
        tokenizer=tokenizer,
    )

    # curate dataset
    mask_token_id = tokenizer.mask_token_id()
    dataset = AtacformerMLMDataset(
        "/Users/nathanleroy/Desktop/gtoks", mask_token_id=mask_token_id, vocab_size=len(tokenizer)
    )
    collator = AtacformerMLMCollator(model.tokenizer.padding_token_id())
    dataloader = DataLoader(
        dataset,
        batch_size=2,
        num_workers=1,
        collate_fn=collator,
    )

    # make adapter and trainer
    adapter = MLMAdapter(model)
    trainer = L.Trainer(
        max_epochs=3,
    )
    trainer.fit(adapter, train_dataloaders=dataloader)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from abc import ABC

import numpy as np
from hmmlearn import hmm

from .custom_distribution import NBHMM, BetaHMM


class Model(ABC):
    """Abstract class of HMM models"""

    def __init__(self, trans_matrix, init_para, para, save_matrix):
        """

        :param array trans_matrix: transition matrix
        :param str init_para: parameters that need to be initialized before training
        :param str para: parameters that are update during training
        :param bool save_matrix: whether to save to a file transition matrix
        """
        self.init_para = init_para
        self.para = para
        self.state_no = 4
        self.save_matrix = save_matrix
        self.trans_matrix = trans_matrix
        self.start_matrix = np.array([0.01, 0.01, 0.97, 0.01])

    def save_tras(self, out_folder):
        np.savetxt(os.path.join(out_folder, "trans_matrix.csv"), self.trans_matrix)


class PoissonModel(Model):
    """HMM model with Poisson emissions"""

    def __init__(
        self,
        trans_matrix,
        lambdas_matrix,
        init_para="",
        para="",
        save_matrix=True,
        out_folder="",
    ):
        """

        :param lambdas_matrix: Lambda values for emission probabilities
        :param out_folder: folder to which save parameter matrix
        """
        Model.__init__(
            self,
            trans_matrix=trans_matrix,
            init_para=init_para,
            para=para,
            save_matrix=save_matrix,
        )
        self.lambdas_matrix = lambdas_matrix
        if save_matrix:
            super().save_tras(out_folder)
            np.savetxt(
                os.path.join(out_folder, "lambdas_matrix.csv"),
                self.lambdas_matrix,
            )

        self.model = hmm.PoissonHMM(
            n_components=self.state_no,
            verbose=True,
            init_params=self.init_para,
            params=self.para,
        )

        if "t" not in self.init_para:
            self.model.transmat_ = self.trans_matrix
        if "l" not in self.init_para:
            self.model.lambdas_ = self.lambdas_matrix
        self.model.startprob_ = self.start_matrix


class GaussianModel(Model):
    def __init__(
        self,
        trans_matrix,
        means_matrix,
        covars_matrix,
        init_para="",
        para="",
        save_matrix=True,
        out_folder="",
    ):
        """

        :param means_matrix: mean values for emission probabilities
        :param covars_matrix: covariance values for emission probabilities
        :param out_folder: folder to which save parameter matrix
        """
        Model.__init__(
            self,
            trans_matrix=trans_matrix,
            init_para=init_para,
            para=para,
            save_matrix=save_matrix,
        )
        self.means_matrix = means_matrix
        self.covars_matrix = covars_matrix
        if save_matrix:
            super().save_tras(out_folder)
            np.savetxt(
                os.path.join(out_folder, "covars_matrix.csv"),
                self.covars_matrix,
            )
            np.savetxt(os.path.join(out_folder, "means_matrix.csv"), self.means_matrix)

        self.model = hmm.GaussianHMM(
            n_components=self.state_no,
            verbose=True,
            init_params=self.init_para,
            params=self.para,
        )

        if "t" not in self.init_para:
            self.model.transmat_ = self.trans_matrix
        if "m" not in self.init_para:
            self.model.means_ = self.means_matrix
        if "c" not in self.init_para:
            self.model.covars_ = self.covars_matrix
        self.model.startprob_ = self.start_matrix


class NBModel(Model):
    def __init__(
        self,
        trans_matrix,
        failures_matrix,
        prob_matrix,
        init_para="",
        para="",
        save_matrix=True,
        out_folder="",
    ):
        """

        :param failures_matrix: number of failures for emission probabilities
        :param prob_matrix: success probability for emission probabilities
        :param out_folder: folder to which save parameter matrix
        """
        Model.__init__(
            self,
            trans_matrix=trans_matrix,
            init_para=init_para,
            para=para,
            save_matrix=save_matrix,
        )
        self.failures_matrix = failures_matrix
        self.prob_matrix = prob_matrix
        if save_matrix:
            super().save_tras(out_folder)
            np.savetxt(
                os.path.join(failures_matrix, "failures_matrix.csv"),
                self.covars_matrix,
            )
            np.savetxt(os.path.join(out_folder, "prob_matrix.csv"), self.prob_matrix)

        self.model = NBHMM(
            n_components=self.state_no,
            verbose=True,
            init_params=self.init_para,
            params=self.para,
        )

        self.model.transmat_ = self.trans_matrix
        self.model.failures_ = self.failures_matrix
        self.model.prob_ = self.prob_matrix
        self.model.startprob_ = self.start_matrix


class BetaModel(Model):
    def __init__(
        self,
        trans_matrix,
        alfa_matrix,
        beta_matrix,
        init_para="",
        para="",
        save_matrix=True,
        out_folder="",
    ):
        """

        :param alfa_matrix: alfa values for emission probabilities
        :param beta_matrix: beta values for emission probabilities
        :param out_folder: folder to which save parameter matrix
        """
        Model.__init__(
            self,
            trans_matrix=trans_matrix,
            init_para=init_para,
            para=para,
            save_matrix=save_matrix,
        )
        self.alfa_matrix = alfa_matrix
        self.beta_matrix = beta_matrix
        if save_matrix:
            super().save_tras(out_folder)
            np.savetxt(os.path.join(out_folder, "alfa_matrix.csv"), self.alfa_matrix)
            np.savetxt(os.path.join(out_folder, "beta_matrix.csv"), self.beta_matrix)

        self.model = BetaHMM(
            n_components=self.state_no,
            verbose=True,
            init_params=self.init_para,
            params=self.para,
        )

        self.model.transmat_ = self.trans_matrix
        self.model.alfa_ = self.alfa_matrix
        self.model.beta_ = self.beta_matrix
        self.model.startprob_ = self.start_matrix



import os
from glob import glob
from math import ceil
from typing import List, Tuple

import torch
from gtars.utils import read_tokens_from_gtok
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

from .const import KEEP_RATE, MASK_RATE, REPLACE_WITH_MASK_RATE, REPLACE_WITH_RANDOM_RATE


class AtacformerMLMCollator:
    """
    Collator for the MLM dataset. This will pad the tokens, masked_tokens, and mask_ids
    """

    def __init__(self, padding_token: int):
        self.padding_token = padding_token

    def __call__(
        self, batch: List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Collate function for the MLM dataset. This should take a batch of
        (tokens, masked_tokens, mask_ids) and return a tuple of (tokens, masked_tokens, mask_ids) that are padded

        :param list[tuple[torch.Tensor, torch.Tensor, torch.Tensor]] batch: Batch of (tokens, masked_tokens, mask_ids)
        :param int padding_token: Token to use for padding
        """
        tokens, masked_tokens, mask_ids = zip(*batch)

        # pad the tokens
        tokens = pad_sequence(tokens, batch_first=True, padding_value=self.padding_token)
        masked_tokens = pad_sequence(
            masked_tokens, batch_first=True, padding_value=self.padding_token
        )
        mask_ids = pad_sequence(mask_ids, batch_first=True, padding_value=self.padding_token)

        attention_mask = (tokens != self.padding_token).float()

        return tokens, masked_tokens, mask_ids, attention_mask


class AtacformerMLMDataset(Dataset):
    def __init__(
        self,
        data: str,
        mask_token_id: int,
        vocab_size: int,
        mask_rate: float = MASK_RATE,
        random_seed: int = 42,
        shuffle: bool = True,
        context_size: int = 2048,
        seed: int = 42,
    ):
        """
        Initialize the MLM dataset. This is heavily based on the MLM dataset
        proposed in the original BERT paper (https://arxiv.org/abs/1810.04805)

        :param str data: Path to the dataset. This should be a file of .gtok files
        :param int mask_token_id: ID of the mask token
        :param float mask_rate: Probability of masking a token
        :param int vocab_size: Size of the vocabulary
        :param int random_seed: Random seed to use
        :param bool shuffle: Whether to shuffle the data
        """
        self.data = data
        self.mask_rate = mask_rate
        self.mask_token_id = mask_token_id
        self.vocab_size = vocab_size
        self.random_seed = random_seed
        self.shuffle = shuffle
        self.context_size = context_size

        # get list of all files
        self.files = glob(os.path.join(data, "*.gtok"), recursive=True)
        if len(self.files) == 0:
            # try recursive
            self.files = glob(os.path.join(data, "**/*.gtok"), recursive=True)

        self.probs = torch.tensor([REPLACE_WITH_MASK_RATE, REPLACE_WITH_RANDOM_RATE, KEEP_RATE])

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        This should return a tuple of (tokens, masked_tokens, mask_ids).
        """
        # load the data into memory
        tokens = torch.tensor(read_tokens_from_gtok(self.files[idx]))

        # reduce the tokens to the context size
        # randomly sample self.context_size tokens from the tokens
        # but dont just slice it.... actually just
        # pick self.context_points without replacement
        if tokens.shape[0] > self.context_size:
            indices = torch.multinomial(
                torch.ones(tokens.shape[0]), self.context_size, replacement=False
            )
            tokens = tokens[indices]

        masked_tokens = tokens.clone()

        # select the tokens to mask
        #   -- each token has an **equal probability** of being masked ( i.e. torch.ones(tokens.shape[0]))
        #   -- we sample a certain percentage of tokens to mask (in BERT this is 15%) ( i.e. ceil(tokens.shape[0] * self.mask_rate) )
        #   -- a token can't be masked more than once (replacement=False)
        mask_ids = torch.multinomial(
            torch.ones(tokens.shape[0]), ceil(tokens.shape[0] * self.mask_rate), replacement=False
        )

        # perform the actual masking. there are three possible outcomes:
        #   1. mask the token
        #   2. replace the token with a random token
        #   3. keep the token the same as it was
        # each outcome has a different probability of happening (`REPLACE_WITH_MASK_RATE`, `REPLACE_WITH_RANDOM_RATE`, `KEEP_RATE` represented in `self.probs`)
        #   therefore, we need a decision made for each token selected to be masked. this means we set num_samples = mask_ids.shape[0] (total number of
        #   tokens selected to mask). we need to sample with replacement (replacement=True) because the same outcome can occur multiple times. e.g. multiple
        #   tokens can be replaced with the mask token.
        random_vals = torch.multinomial(self.probs, mask_ids.shape[0], replacement=True)

        # now actually mask the tokens. recall that we defined the distribution as:
        #   `[REPLACE_WITH_MASK_RATE, REPLACE_WITH_RANDOM_RATE, KEEP_RATE]`
        #   so if the random value is:
        #   - 0, we *mask the token*,
        #   - 1, we *replace it with a random token*
        #   - 2, we *keep it the same*
        mask_token_indices = mask_ids[random_vals == 0]
        replace_random_indices = mask_ids[random_vals == 1]
        # why is there no need to do anything for random_vals == 2? because we're keeping the token the same...

        # perform the masking
        masked_tokens[mask_token_indices] = self.mask_token_id
        masked_tokens[replace_random_indices] = torch.randint(
            self.vocab_size, (replace_random_indices.shape[0],)
        )

        # when training we need to pass the masked tokens to the model
        # but we also need to know which tokens were masked so we can calculate the loss
        # and we need to know what those tokens were replaced with so we can calculate the loss
        return tokens, masked_tokens, mask_ids

    def __str__(self):
        return f"AtacformerMLMDataset({len(self)} files)"

    def __repr__(self):
        return f"AtacformerMLMDataset({len(self)} files)"


class AtacformerCellTypeFineTuningCollator:
    """
    Collator for the cell type fine-tuning dataset. This will pad the tokens, labels, and attention_mask
    """

    def __init__(self, padding_token: int):
        self.padding_token = padding_token

    def __call__(
        self, batch: List[Tuple[torch.Tensor, torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Collate function for the cell type fine-tuning dataset. This should take a batch of
        (tokens, labels) and return a tuple of (tokens, labels) that are padded

        :param list[tuple[torch.Tensor, torch.Tensor]] batch: Batch of (tokens, labels)
        :param int padding_token: Token to use for padding
        """
        cell1, cell2, labels = zip(*batch)

        # pad the tokens
        cell1 = pad_sequence(cell1, batch_first=True, padding_value=self.padding_token)
        cell2 = pad_sequence(cell2, batch_first=True, padding_value=self.padding_token)

        attention_mask1 = (cell1 != self.padding_token).float()
        attention_mask2 = (cell2 != self.padding_token).float()

        labels = torch.tensor(labels)

        return cell1, cell2, labels, attention_mask1, attention_mask2


class AtacformerCellTypeFineTuningDataset(Dataset):
    def __init__(
        self,
        file_path: str,
        context_size: int = 2048,
        seed: int = 42,
    ):
        """
        Initialize the cell type fine-tuning dataset.

        :param str file_path: Path to the file that defines the dataset
        """
        # check file and not directory
        if os.path.isdir(file_path):
            raise ValueError(f"Expected a file, got a directory: {file_path}")

        if not os.path.exists(file_path):
            raise ValueError(f"File does not exist: {file_path}")

        # init params
        self.file_path = file_path
        self.root_dir = os.path.dirname(self.file_path)
        self.context_size = context_size
        self.seed = seed

        # read the file, line by line
        # format is a\tb\label
        self.pairs = []
        with open(file_path, "r") as f:
            for line in f:
                barcode1, barcode2, label, cell_type1, cell_type2 = line.strip().split("\t")
                barcode1_path = os.path.join(self.root_dir, barcode1)
                barcode2_path = os.path.join(self.root_dir, barcode2)
                self.pairs.append((barcode1_path, barcode2_path, int(label)))

        self.max_label = max([pair[2] for pair in self.pairs])

    def __len__(self):
        return len(self.pairs)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        This should return a tuple of (tokens, label).
        """
        barcode1_path, barcode2_path, label = self.pairs[idx]

        # load the data into memory
        cell1 = torch.tensor(read_tokens_from_gtok(barcode1_path + ".gtok"))
        cell2 = torch.tensor(read_tokens_from_gtok(barcode2_path + ".gtok"))

        # reduce the tokens to the context size
        # randomly sample self.context_size tokens from the tokens
        if cell1.shape[0] > self.context_size:
            indices = torch.multinomial(
                torch.ones(cell1.shape[0]), self.context_size, replacement=False
            )
            cell1 = cell1[indices]

        if cell2.shape[0] > self.context_size:
            indices = torch.multinomial(
                torch.ones(cell2.shape[0]), self.context_size, replacement=False
            )
            cell2 = cell2[indices]

        label = torch.tensor(label)

        return cell1, cell2, label

    def __str__(self):
        return f"AtacformerCellTypeFineTuningDataset({len(self)} files)"

    def __repr__(self):
        return f"AtacformerCellTypeFineTuningDataset({len(self)} files)"



import argparse
import glob
import multiprocessing as mp
import os
import pickle
import random
import time

import numpy as np
import sklearn.neural_network as nn
from gensim.models import Word2Vec
from sklearn.compose import TransformedTargetRegressor
from sklearn.model_selection import KFold, cross_val_score, train_test_split
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from ..utils import timer_func
from .utils import cosine_distance, genome_distance, load_genomic_embeddings


def get_rct_score(
    path: str,
    embed_type: str,
    bin_path: str,
    out_dim: int = -1,
    cv_num: int = 5,
    seed: int = 42,
    num_workers: int = 10,
) -> float:
    """Runs the RCT on a model.

    Args:
        path (str): The path to a model.
        embed_type (str): The type of the model: "region2vec" or "base".
        bin_path (str): The path to a binary embedding model.
        out_dim (int, optional): Output dimension of a prediction. Defaults to
            -1, and out_dim is the same as the dimension of a binary embedding.
        cv_num (int, optional): Number of folds in cross-validation. Defaults
            to 5.
        seed (int, optional): Random seed. Defaults to 42.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Returns:
        float: the RCT score.
    """
    embed_rep, vocab = load_genomic_embeddings(path, embed_type)
    embed_bin, vocab_bin = load_genomic_embeddings(bin_path, "base")
    region2idx = {r: i for i, r in enumerate(vocab)}
    region2idx_bin = {r: i for i, r in enumerate(vocab_bin)}
    # align embed_bin with embed_rep
    if out_dim <= 0:
        embed_bin = np.array([embed_bin[region2idx_bin[v]] for v in vocab])
    else:
        bin_dim = embed_bin.shape[1]
        out_dim = min(bin_dim, out_dim)
        sel_dims = np.random.choice(bin_dim, out_dim)
        embed_bin = np.array([embed_bin[region2idx_bin[v]][sel_dims] for v in vocab])

    regressor = nn.MLPRegressor(
        hidden_layer_sizes=(200),
        activation="relu",
        solver="adam",
        alpha=0.0001,  # regularizer strength
        batch_size="auto",
        learning_rate_init=0.001,
        max_iter=200,
        shuffle=True,
        random_state=seed,
        tol=0.0001,
        verbose=False,
        early_stopping=False,
        validation_fraction=0.1,
        n_iter_no_change=10,
    )
    model_in = make_pipeline(StandardScaler(), regressor)
    model = TransformedTargetRegressor(regressor=model_in, transformer=StandardScaler())

    kf = KFold(n_splits=cv_num, shuffle=True, random_state=seed)
    if num_workers > cv_num:
        num_workers = cv_num
    score = cross_val_score(model, embed_rep, embed_bin, cv=kf, n_jobs=num_workers, verbose=0)
    return score.mean()


def reconstruction_batch(
    batch: list[tuple[str, str, str]],
    cv_num: int,
    out_dim: int = -1,
    seed: int = 42,
    save_path: str = None,
    num_workers: int = 10,
) -> list[tuple[str, float]]:
    """Runs the RCT on a batch of models.

    Args:
        batch (list[tuple[str, str, str]]): A list of (model path, model type,
            binary embedding path) tuples. Model type could be "region2vec" or
            "base".
        cv_num (int): Number of folds in cross-validation. Defaults
            to 5.
        out_dim (int, optional): Output dimension of a prediction. Defaults to
            -1, and out_dim is the same as the dimension of a binary embedding.
        seed (int, optional): Random seed. Defaults to 42.
        save_path (str, optional): Save the results to save_path. Defaults to
            None.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Returns:
        list[tuple[str, float]]: A list of (model path, RCT score) tuples.
    """
    rct_arr = []
    for path, embed_type, bin_path in batch:
        score = get_rct_score(path, embed_type, bin_path, out_dim, cv_num, seed, num_workers)
        rct_arr.append((path, score))
    if save_path:
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        with open(save_path, "wb") as f:
            pickle.dump(rct_arr, f)
    return rct_arr


def rct_eval(
    batch: list[tuple[str, str, str]],
    num_runs: int = 5,
    cv_num: int = 5,
    out_dim: int = -1,
    save_folder: str = None,
    num_workers: int = 10,
) -> list[tuple[str, list[float]]]:
    """Runs the RCT on a batch of models for multiple times.

    Args:
        batch (list[tuple[str, str, str]]): A list of (model path, model type,
            binary embedding path) tuples. Model type could be "region2vec" or
            "base".
        num_runs (int, optional): Number of runs. Defaults to 5.
        cv_num (int, optional): Number of folds in cross-validation.
            Defaults to 5.
        out_dim (int, optional): Output dimension of a prediction. Defaults to
            -1, and out_dim is the same as the dimension of a binary embedding.
        save_folder (str, optional): Folder to save the results from each run.
            Defaults to None.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Returns:
        list[tuple[str, list[float]]]: A list of (model path, RCT scores from
            num_runs) tuples.
    """
    results_seeds = []
    for seed in range(num_runs):
        print(f"----------------Run {seed}----------------")
        save_path = os.path.join(save_folder, f"rct_eval_seed{seed}") if save_folder else None
        result_list = reconstruction_batch(batch, cv_num, out_dim, seed, save_path, num_workers)
        results_seeds.append(result_list)

    rct_res = [[] for i in range(len(batch))]
    for results in results_seeds:
        for i, res in enumerate(results):
            rct_res[i].append(res[1])
            assert res[0] == batch[i][0], "key == batch[i][0]"
    mean_rct = [np.array(r).mean() for r in rct_res]
    std_rct = [np.array(r).std() for r in rct_res]
    models = [t[0] for t in batch]
    for i in range(len(mean_rct)):
        print(f"{batch[i][0]}\n RCT (std): {mean_rct[i]:.4f} ({std_rct[i]:.4f}) \n")
    rct_arr = [(batch[i][0], rct_res[i]) for i in range(len(batch))]
    return rct_arr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np
from hmmlearn.base import BaseHMM
from scipy.stats import beta, nbinom
from sklearn.utils import check_random_state


def _check_and_set_n_features(model, X):
    _, n_features = X.shape
    if hasattr(model, "n_features"):
        if model.n_features != n_features:
            raise ValueError(
                f"Unexpected number of dimensions, got {n_features} but "
                f"expected {model.n_features}"
            )
    else:
        model.n_features = n_features


class NBHMM(BaseHMM):
    def __init__(
        self,
        n_components=1,
        startprob_prior=1.0,
        transmat_prior=1.0,
        failures_prior=0.0,
        prob_prior=0.0,
        failures_weight=0.0,
        prob_weight=0.0,
        algorithm="viterbi",
        random_state=None,
        n_iter=10,
        tol=1e-2,
        verbose=False,
        params="strp",
        init_params="strp",
        implementation="log",
    ):
        BaseHMM.__init__(
            self,
            n_components,
            startprob_prior=startprob_prior,
            transmat_prior=transmat_prior,
            algorithm=algorithm,
            random_state=random_state,
            n_iter=n_iter,
            tol=tol,
            verbose=verbose,
            params=params,
            init_params=init_params,
            implementation=implementation,
        )
        self.failures_prior = failures_prior
        self.failures_weight = failures_weight
        self.prob_prior = prob_prior
        self.prob_weight = prob_weight

    def _init(self, X):
        _check_and_set_n_features(self, X)
        super()._init(X)
        self.random_state = check_random_state(self.random_state)

        mean_X = X.mean()
        var_X = X.var()

        if self._needs_init("p", "prob_"):
            # initialize with method of moments based on X
            raise NotImplementedError

        if self._needs_init("r", "failures_"):
            # initialize with method of moments based on X
            raise NotImplementedError

    def _get_n_fit_scalars_per_param(self):
        nc = self.n_components
        nf = self.n_features
        return {
            "s": nc - 1,
            "t": nc * (nc - 1),
            "p": nc * nf,
            "r": nc * nf,
        }

    def _check(self):
        super()._check()

        self.prob_ = np.atleast_2d(self.prob_)
        self.failures_ = np.atleast_2d(self.failures_)
        n_features = getattr(self, "n_features", self.prob_.shape[1])
        if self.prob_.shape != (self.n_components, n_features):
            raise ValueError("proba_ must have shape (n_components, n_features)")
        if self.failures_.shape != (self.n_components, n_features):
            raise ValueError("failures_ must have shape (n_components, n_features)")
        if self.failures_.shape != self.prob_.shape:
            raise ValueError(
                "failures_ and proba_ must have the same shape (n_components, n_features)"
            )
        self.n_features = n_features

    def _generate_sample_from_state(self, state, random_state):
        return random_state.negative_binomial(self.failures_[state], self.prob_[state])

    def _compute_log_likelihood(self, X):
        return np.array(
            [
                np.sum(nbinom.logpmf(X, failures, proba), axis=1)
                for failures, proba in zip(self.failures_, self.prob_)
            ]
        ).T

    def _compute_likelihood(self, X):
        return np.array(
            [
                np.prod(nbinom.pmf(X, failures, proba), axis=1)
                for failures, proba in zip(self.failures_, self.prob_)
            ]
        ).T

    def _initialize_sufficient_statistics(self):
        stats = super()._initialize_sufficient_statistics()
        stats["post"] = np.zeros(self.n_components)
        stats["obs"] = np.zeros((self.n_components, self.n_features))
        return stats

    def _accumulate_sufficient_statistics(
        self, stats, obs, lattice, posteriors, fwdlattice, bwdlattice
    ):
        super()._accumulate_sufficient_statistics(
            stats, obs, lattice, posteriors, fwdlattice, bwdlattice
        )
        raise NotImplementedError

    def _do_mstep(self, stats):
        super()._do_mstep(stats)
        raise NotImplementedError


class BetaHMM(BaseHMM):
    def __init__(
        self,
        n_components=1,
        startprob_prior=1.0,
        transmat_prior=1.0,
        alfa_prior=0.0,
        beta_prior=0.0,
        alfa_weight=0.0,
        beta_weight=0.0,
        algorithm="viterbi",
        random_state=None,
        n_iter=10,
        tol=1e-2,
        verbose=False,
        params="strp",
        init_params="strp",
        implementation="log",
    ):
        BaseHMM.__init__(
            self,
            n_components,
            startprob_prior=startprob_prior,
            transmat_prior=transmat_prior,
            algorithm=algorithm,
            random_state=random_state,
            n_iter=n_iter,
            tol=tol,
            verbose=verbose,
            params=params,
            init_params=init_params,
            implementation=implementation,
        )
        self.alfa_prior = alfa_prior
        self.alfa_weight = alfa_weight
        self.beta_prior = beta_prior
        self.beta_weight = beta_weight

    def _init(self, X):
        _check_and_set_n_features(self, X)
        super()._init(X)
        self.random_state = check_random_state(self.random_state)

        if self._needs_init("a", "alfa_"):
            raise NotImplementedError

        if self._needs_init("b", "beta_"):
            raise NotImplementedError

    def _get_n_fit_scalars_per_param(self):
        nc = self.n_components
        nf = self.n_features
        return {
            "s": nc - 1,
            "t": nc * (nc - 1),
            "a": nc * nf,
            "b": nc * nf,
        }

    def _check(self):
        super()._check()

        self.alfa_ = np.atleast_2d(self.alfa_)
        self.beta_ = np.atleast_2d(self.beta_)
        n_features = getattr(self, "n_features", self.alfa_.shape[1])
        if self.alfa_.shape != (self.n_components, n_features):
            raise ValueError("alfa_ must have shape (n_components, n_features)")
        if self.beta_.shape != (self.n_components, n_features):
            raise ValueError("beta_ must have shape (n_components, n_features)")
        if self.alfa_.shape != self.beta_.shape:
            raise ValueError("alfa_ and beta_ must have the same shape (n_components, n_features)")
        self.n_features = n_features

    def _generate_sample_from_state(self, state, random_state):
        return random_state.beta(self.alfa_[state], self.beta_[state])

    def _compute_log_likelihood(self, X):
        return np.array(
            [np.sum(beta.logpdf(X, a, b), axis=1) for a, b in zip(self.alfa_, self.beta_)]
        ).T

    def _compute_likelihood(self, X):
        return np.array(
            [np.prod(beta.pdf(X, a, b), axis=1) for a, b in zip(self.alfa_, self.beta_)]
        ).T

    def _initialize_sufficient_statistics(self):
        stats = super()._initialize_sufficient_statistics()
        raise NotImplementedError

    def _accumulate_sufficient_statistics(
        self, stats, obs, lattice, posteriors, fwdlattice, bwdlattice
    ):
        super()._accumulate_sufficient_statistics(
            stats, obs, lattice, posteriors, fwdlattice, bwdlattice
        )
        raise NotImplementedError

    def _do_mstep(self, stats):
        super()._do_mstep(stats)
        raise NotImplementedError



import argparse
import glob
import os
import pickle
import random
import time
from typing import List

import numpy as np

from geniml.region2vec import utils


class BEDDataset:
    """Wraps a set of BED files in a BEDDataset object.

    Stores the information of a set of BED files.
    Generates a new dataset with regions shuffled in BED files.

    Attributes:
        filename_list: A list of BED file names.
        nfiles: The number of BED files in the BEDdataset object.
    """

    def __init__(self, file_list: str) -> None:
        """Initializes a BEDDataset object.

        Args:
            file_list (str): A file storing a list of BED file names that
                should be included in the dataset.
        """
        self.filename_list = []
        with open(file_list, "r") as f:
            for idx, line in enumerate(f):
                filename = line.strip()
                self.filename_list.append(filename)

        self.nfiles = len(self.filename_list)

    def regions2sentences_sampling(self, src_path: str, dst_path: str) -> None:
        """Constructs a sentence by sampling regions from a BED file.

        Each region in a BED file has a probability. Constructs a sentence by
        sampling regions in a BED file based on their probabilities.

        Args:
            src_path (str): The folder where BED files reside.
            dst_path (str): The destination file that stores all the generated
                BED files; each line has regions sampled from a BED file.
        """
        with open(dst_fname, "w") as fout:
            for fname in self.filename_list:
                src_fname = os.path.join(src_path, fname)
                sentence = []
                probs = []
                with open(src_fname, "r") as f:
                    for line in f:
                        elements = line.strip().split("\t")
                        word = elements[0].strip()
                        sentence.append(word)
                        probs.append(float(elements[-2].strip()))
                probs = np.array(probs)
                probs = probs / probs.sum()
                sentence = np.array(sentence)

                sampled_sentence = np.random.choice(sentence, len(probs), p=probs)
                # sampled_sentence = list(set(sampled_sentence))
                sampled_sentence = sampled_sentence.tolist()
                str_sent = " ".join(sampled_sentence)

                fout.write(str_sent)
                fout.write("\n")

    def regions2sentences(self, src_path: str, dst_path: str) -> None:
        """Concatenates all regions in a BED file randomly into a sentence.

        This functions is called in the hard tokenization mode.

        Args:
            src_path (str): The folder where BED files reside.
            dst_path (str): The destination file that stores all the generated
                BED files; each line has all the regions from a BED file.
        """
        with open(dst_path, "w") as f_out:
            for fname in self.filename_list:
                src_fname = os.path.join(src_path, fname)
                sentence = []
                with open(src_fname, "r") as f:
                    for line in f:
                        elements = line.strip().split("\t")[0:3]
                        chr_name = elements[0].strip()
                        start = elements[1].strip()
                        end = elements[2].strip()
                        word = chr_name + ":" + start + "-" + end
                        sentence.append(word)
                random.shuffle(sentence)  # shuffle the regions in the sentence
                str_sent = " ".join(sentence)
                f_out.write(str_sent)
                f_out.write("\n")


class MatrixDataset:
    """Wraps the binary representation of BED files into a MatrixDataset.

    Stores the information of a set of BED files.
    Generates a new dataset with regions shuffled in BED files.
    """

    def __init__(self, matrix: List[List[int]]):
        """Initializes a MatrixDataset object with matrix.

        Args:
            matrix (list[list[int]]): The binary representation of BED files.
                Each row represents a BED file. Each column denotes a region.
                Each element denotes the presence (1) or absence (0) of a
                region in a BED file.
        """
        self.mat = [[] for i in range(len(matrix))]
        for i in range(len(matrix)):
            for j in range(len(matrix[i])):
                if matrix[i][j] != 0:
                    self.mat[i].append(j)

    def regions2sentences(self, dst_path: str) -> None:
        """Concatenates all regions in a BED file randomly into a sentence.

        This functions is called in the hard tokenization mode.

        Args:
            dst_path (str): The destination file that stores all the generated
                BED files; each line has all the regions from a BED file.
        """
        with open(dst_path, "w") as f_out:
            for i in range(len(self.mat)):
                sentence = []
                for j in range(len(self.mat[i])):
                    sentence.append(self.mat[i][j])
                if len(sentence) > 0:
                    random.shuffle(sentence)  # shuffle the regions in the sentence
                    str_sent = " ".join(sentence)
                    f_out.write(str_sent)
                    f_out.write("\n")


def main(args: argparse.Namespace) -> None:
    """Generates shuffled datasets.

    Called internally as a subprocess by region2vec in main.py.

    Args:
        args (argparse.Namespace): See the definition of ArgumentParser.
    """
    DATA_FOLDER = os.path.join(args.save_dir, "shuffled_datasets")
    os.makedirs(DATA_FOLDER, exist_ok=True)
    src_path = args.tokenization_folder
    worker_id = args.worker_id
    random.seed(worker_id)
    np.random.seed(worker_id)
    if args.data_type == "files":
        dataset = BEDDataset(args.file_list)
    else:
        with open(args.mat_path, "rb") as f:
            matrix = pickle.load(f)
        dataset = MatrixDataset(matrix)
    pool = args.pool
    utils.log(f"[{worker_id}] Creating shuffled datasets in \033[93m{DATA_FOLDER}\033[00m")

    for i in range(pool):
        name_used = os.path.join(DATA_FOLDER, f"pool{worker_id}-{i}used")
        name_using = os.path.join(DATA_FOLDER, f"pool{worker_id}-{i}using")
        name_creating = os.path.join(DATA_FOLDER, f"pool{worker_id}-{i}creating")
        name = os.path.join(DATA_FOLDER, f"pool{worker_id}-{i}")
        if os.path.exists(name_using):
            print("File exists")
            return
        if os.path.exists(name_used):
            print("File exists")
            return
        if os.path.exists(name):
            print("File exists")
            return
        if os.path.exists(name_creating):
            print("File exists")
            return
        # create an empty file
        with open(name_used, "w") as f:
            pass

    num_created = 0
    while True:
        if num_created == args.number:
            break
        # determine whether to create a new dataset
        files = glob.glob(os.path.join(DATA_FOLDER, f"pool{worker_id}*used"))
        if len(files) == 0:
            time.sleep(1)  # wait for 10 seconds
            # print('Waiting for the data to be consumed',end="\r")
        else:
            # delete the used dataset and generate a new dataset in the same folder
            sel_file = files[random.randint(0, len(files) - 1)]
            fname = sel_file.split("/")[-1][:-4]
            os.system(f"rm -f {sel_file}")  # delete the dataset
            dpath = os.path.join(DATA_FOLDER, fname + "creating")
            with open(dpath, "w") as f:
                pass
            if args.tokenization_mode == "hard":
                dataset.regions2sentences(src_path, dpath)
            else:
                dataset.regions2sentences_sampling(src_path, dpath)

            num_created += 1
            # print('[',datetime.datetime.now(),']',' Created %dth dataset' % num_created)
            dst_name = os.path.join(DATA_FOLDER, fname)
            os.rename(dpath, dst_name)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Sentence Generation")
    parser.add_argument("--file-list", help="path to a file list")
    parser.add_argument("--tokenization-mode", help="tokenization mode")
    parser.add_argument(
        "--tokenization-folder",
        help="path to the folder that saves tokenized regions",
    )
    parser.add_argument("--save-dir", help="parent folder to generated shuffled datasets")
    parser.add_argument(
        "--pool",
        type=int,
        default=3,
        help="maximum number of shuffled datasets before consuming one",
    )
    parser.add_argument(
        "--worker-id",
        type=int,
        default=0,
        help="used in the parallel mode",
    )
    parser.add_argument(
        "--number",
        type=int,
        default=1000,
        help="number of shuffling the whole dataset",
    )

    args = parser.parse_args()

    main(args)



import lightning as L
import pytest
import scanpy as sc
import torch
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

from geniml.region2vec.main import Region2Vec, Region2VecExModel
from geniml.tokenization.main import AnnDataTokenizer

# from geniml.training import CellTypeFineTuneAdapter
# from geniml.training.utils import (
#     FineTuningDataset,
#     collate_finetuning_batch,
#     generate_fine_tuning_dataset,
# )


@pytest.mark.skip("Too slow for CI/CD. Mostly a development tool anyways.")
class TestFinetuining:

    def test_generate_finetuning_dataset(
        self,
    ):
        t = AnnDataTokenizer("tests/data/universe.bed")
        adata = sc.read_h5ad("tests/data/pbmc_hg38.h5ad")

        pos, neg, pos_labels, neg_labels = generate_fine_tuning_dataset(
            adata, t, negative_ratio=1.0, sample_size=2
        )

        # total positive pairs should be equal to total negative pairs
        # total positive pairs will be equal to sum([n*(n - 1) for n in adata.obs.groupby("cell_type").size()])
        # total negative pairs only equals number of positive pairs when negative_ratio=1.0

        assert len(pos) == len(neg)
        assert len(pos) == len(pos_labels)
        assert len(neg) == len(neg_labels)
        # not sure why the below doesnt work right now
        # assert len(pos) == sum([(n * (n - 1)) for n in adata.obs.groupby("cell_type").size()])

    @pytest.mark.skip("Too slow for CI/CD. Mostly a development tool anyways.")
    def test_init_celltype_adapter(
        self,
    ):
        model = Region2VecExModel(
            tokenizer="tests/data/universe.bed",
        )
        adapter = CellTypeFineTuneAdapter(model)
        assert adapter is not None
        assert isinstance(adapter.r2v_model, Region2Vec)
        assert adapter.r2v_model.projection.num_embeddings == len(model.tokenizer)

    @pytest.mark.skip("Too slow for CI/CD. Mostly a development tool anyways.")
    def test_train_with_adapter(
        self,
    ):
        # make models
        model = Region2VecExModel(
            tokenizer="tests/data/universe.bed",
        )
        adapter = CellTypeFineTuneAdapter(model)

        # load data
        data = sc.read_h5ad("tests/data/pbmc_hg38.h5ad")
        pos_pairs, neg_pairs, pos_labels, neg_labels = generate_fine_tuning_dataset(
            data, model.tokenizer, seed=42, negative_ratio=1.0, sample_size=1_000
        )

        # combine the positive and negative pairs
        pairs = pos_pairs + neg_pairs
        labels = pos_labels + neg_labels

        # get the pad token id
        pad_token_id = model.tokenizer.padding_token_id()

        train_pairs, test_pairs, Y_train, Y_test = train_test_split(
            pairs,
            labels,
            train_size=0.8,
            random_state=42,
        )

        batch_size = 32

        # create the datasets
        train_dataloader = DataLoader(
            FineTuningDataset(train_pairs, Y_train),
            batch_size=batch_size,
            shuffle=True,
            collate_fn=lambda x: collate_finetuning_batch(x, pad_token_id),
            # num_workers=multiprocessing.cpu_count() - 2,
        )
        test_dataloader = DataLoader(
            FineTuningDataset(test_pairs, Y_test),
            batch_size=batch_size,
            shuffle=True,
            collate_fn=lambda x: collate_finetuning_batch(x, pad_token_id),
            # num_workers=multiprocessing.cpu_count() - 2,
        )

        trainer = L.Trainer(profiler="simple", min_epochs=3)
        trainer.fit(adapter, train_dataloaders=train_dataloader, val_dataloaders=test_dataloader)

    @pytest.mark.skip("Too slow for CI/CD. Mostly a development tool anyways.")
    def test_train_export(
        self,
    ):
        # make models
        model = Region2VecExModel(
            tokenizer="tests/data/universe.bed",
        )
        adapter = CellTypeFineTuneAdapter(model)

        # load data
        data = sc.read_h5ad("tests/data/pbmc_hg38.h5ad")
        pos_pairs, neg_pairs, pos_labels, neg_labels = generate_fine_tuning_dataset(
            data, model.tokenizer, seed=42, negative_ratio=1.0, sample_size=1_000
        )

        # combine the positive and negative pairs
        pairs = pos_pairs + neg_pairs
        labels = pos_labels + neg_labels

        # get the pad token id
        pad_token_id = model.tokenizer.padding_token_id()

        train_pairs, test_pairs, Y_train, Y_test = train_test_split(
            pairs,
            labels,
            train_size=0.8,
            random_state=42,
        )

        batch_size = 32

        # create the datasets
        train_dataloader = DataLoader(
            FineTuningDataset(train_pairs, Y_train),
            batch_size=batch_size,
            shuffle=True,
            collate_fn=lambda x: collate_finetuning_batch(x, pad_token_id),
            # num_workers=multiprocessing.cpu_count() - 2,
        )
        test_dataloader = DataLoader(
            FineTuningDataset(test_pairs, Y_test),
            batch_size=batch_size,
            shuffle=True,
            collate_fn=lambda x: collate_finetuning_batch(x, pad_token_id),
            # num_workers=multiprocessing.cpu_count() - 2,
        )

        trainer = L.Trainer(min_epochs=3)
        trainer.fit(adapter, train_dataloaders=train_dataloader, val_dataloaders=test_dataloader)

        # get the tensor dfor token 42
        t_before = model._model.projection(torch.tensor([42]))

        # export
        model.export("tests/data/model-tests")

        # load the model
        model = Region2VecExModel.from_pretrained("tests/data/model-tests")

        # get the tensor for token 42
        t_after = model._model.projection(torch.tensor([42]))

        # make sure the tensors are close
        assert torch.allclose(t_before, t_after)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tarfile
import tempfile

import numpy as np
import pyBigWig

from ..utils import read_chromosome_from_bw, timer_func


def model_binomial(folder_in, in_file, chrom, file_out, file_no=None):
    """Create binomial likelihood model
    first column - likelihood of background
    second column - likelihood of coverage"""
    in_file = os.path.join(folder_in, in_file)
    distr_cov = read_chromosome_from_bw(in_file, chrom)
    chrom_size = len(distr_cov)
    no_possible = file_no * len(distr_cov)  # number of possible spots covered
    no_cov = np.sum(distr_cov)  # number of spots covered
    cov_uniq = np.unique(distr_cov).astype(np.uint16)
    prob_array = np.zeros((int(np.max(cov_uniq)) + 1, 2))
    for i in cov_uniq:
        prob_array[i] = [
            np.log10((file_no - i) / (no_possible - no_cov) + 1e-10),
            np.log10((i / no_cov) + 1e-10),
        ]
    header = f"{chrom}_{chrom_size}"
    r = {header: prob_array}
    np.savez_compressed(file_out, **r)


class ChromosomeModel:
    def __init__(self, folder, chrom):
        """
        :param str folder: file with the model
        :param str chrom: of which chromosome is the model
        """
        self.folder = folder
        self.chromosome = chrom
        self.start_file = f"{self.chromosome}_start"
        self.core_file = f"{self.chromosome}_core"
        self.end_file = f"{self.chromosome}_end"
        self.files = {
            "start": self.start_file + ".npz",
            "core": self.core_file + ".npz",
            "end": self.end_file + ".npz",
        }
        self.models = {}

    def __getitem__(self, item):
        return self.models[item]

    def make_model(self, coverage_folder, coverage_prefix, file_no):
        """
        Make a lh model of given chromosome from coverage files
        :param str coverage_folder: path to name with coverage files
        :param str coverage_prefix: prefixed used for making coverage files
        :param int file_no: number of files from which model is being created
        """
        model_binomial(
            coverage_folder,
            f"{coverage_prefix}_start.bw",
            self.chromosome,
            os.path.join(self.folder, self.start_file),
            file_no,
        )
        model_binomial(
            coverage_folder,
            f"{coverage_prefix}_core.bw",
            self.chromosome,
            os.path.join(self.folder, self.core_file),
            file_no,
        )
        model_binomial(
            coverage_folder,
            f"{coverage_prefix}_end.bw",
            self.chromosome,
            os.path.join(self.folder, self.end_file),
            file_no,
        )

    def read(self):
        """
        Read model
        """
        model_folder = tarfile.open(self.folder, "r")
        for f in self.files:
            file = model_folder.extractfile(self.files[f])
            values = np.load(file)
            self.models[f] = values[values.files[0]]
        model_folder.close()

    def read_track(self, track):
        """
        Read specific track from model
        """
        model_folder = tarfile.open(self.folder, "r")
        file = model_folder.extractfile(self.files[track])
        values = np.load(file)
        self.models[track] = values[values.files[0]]
        model_folder.close()


class ModelLH:
    def __init__(self, file):
        """
        Likelihood model class
        :param str file: file containing the model
        """
        self.name = file
        self.chromosomes_list = []
        self.chromosomes_models = {}
        if os.path.exists(self.name):
            if tarfile.is_tarfile(self.name):
                files = tarfile.open(self.name, "r")
                chroms = files.getnames()
                self.chromosomes_list = list(set([i.split("_")[0] for i in chroms]))

    def __getitem__(self, item):
        return self.chromosomes_models[item]

    def make(self, coverage_folder, coverage_prefix, file_no, force=False):
        """
        Make lh model for all chromosomes
        :param str coverage_folder: folder with coverage files
        :param str coverage_prefix: prefixed used for making coverage files
        :param int file_no: number of file from which model is being made
        :param bool force: if overwrite an existing model
        """
        if os.path.exists(self.name):
            if not force:
                print("Model already exists. If you want to overwrite it use force argument")
                return
            else:
                print("Overwriting existing model")
        tar_arch = tarfile.open(self.name, "w")
        temp_dir = tempfile.TemporaryDirectory()
        bw_start = pyBigWig.open(os.path.join(coverage_folder, f"{coverage_prefix}_start.bw"))
        chroms = bw_start.chroms()
        bw_start.close()
        self.chromosomes_list = [i for i in chroms if chroms[i] != 0]
        for c in self.chromosomes_list:
            chrom_model = ChromosomeModel(temp_dir.name, c)
            chrom_model.make_model(
                coverage_folder,
                coverage_prefix,
                file_no,
            )
            for f in chrom_model.files:
                tar_arch.add(
                    os.path.join(temp_dir.name, chrom_model.files[f]),
                    arcname=chrom_model.files[f],
                )
                os.remove(os.path.join(temp_dir.name, chrom_model.files[f]))
        temp_dir.cleanup()
        tar_arch.close()

    def read_chrom(self, chrom):
        """
        Read into model specific chromosome
        """
        self.chromosomes_models[chrom] = ChromosomeModel(self.name, chrom)
        self.chromosomes_models[chrom].read()

    def read_chrom_track(self, chrom, track):
        """
        Read into model specific track for chromosome
        """
        self.chromosomes_models[chrom] = ChromosomeModel(self.name, chrom)
        self.chromosomes_models[chrom].read_track(track)

    def clear_chrom(self, chrom):
        """
        Clear model for given chromosome
        """
        self.chromosomes_models[chrom] = None


@timer_func
def main(model_file, coverage_folder, coverage_prefix, file_no=None, force=False):
    """
    Crate likelihood models for all chromosomes
    :param str model_file: output name
    :param str coverage_folder: folder with coverage files
    :param str coverage_prefix: prefix used for making coverage files
    :param int file_no: number of files used for making coverage tracks
    """
    model = ModelLH(model_file)
    model.make(coverage_folder, coverage_prefix, file_no, force)



import json
import os.path
import pickle
from typing import Dict, List, Union

import hnswlib
import yaml

from ... import _LOGGER

DEP_HNSWLIB = True
# try:
#
#
#
# except ImportError:
#     DEP_HNSWLIB = False
#     _LOGGER.error(
#         "HNSWBackend requires hnswlib. Install hnswlib, or ignore this if you don't need HNSWBackend"
#     )

import numpy as np

from geniml.search.const import (
    DEFAULT_DIM,
    DEFAULT_EF,
    DEFAULT_HNSW_SPACE,
    DEFAULT_INDEX_PATH,
    DEFAULT_M,
)

from ..utils import verify_load_inputs
from .abstract import EmSearchBackend

# if not DEP_HNSWLIB:
#
#     class HNSWBackend(EmSearchBackend):
#         pass
#
# else:


class HNSWBackend(EmSearchBackend):
    """A search backend that uses a local HNSW index to store and search embeddings"""

    # instance variables, should not be class variables

    def __init__(
        self,
        local_index_path: str = DEFAULT_INDEX_PATH,
        payloads: Union[dict, str] = dict(),
        space: str = DEFAULT_HNSW_SPACE,
        dim: int = DEFAULT_DIM,
        ef: int = DEFAULT_EF,
        m: int = DEFAULT_M,
    ):
        """
        Initiate the backend

        :param local_index_path: local path where the index is saved to
        :param space: possible options are l2, cosine or ip
        :param dim: dimension of vectors that will be stored
        :param ef: defines a construction time/accuracy trade-off, higher ef -> more accurate but slower
        :param m: connected with internal dimensionality of the data, higher M -> higher accuracy/run_time
        when ef is fixed
        """
        # super(HNSWBackend, self).__init__()
        # initiate the index
        self.idx = hnswlib.Index(space=space, dim=dim)  # possible options are l2, cosine or ip
        self.idx.init_index(max_elements=0, ef_construction=ef, M=m)

        # load from local index that already store vectors
        if os.path.exists(local_index_path):
            self.idx.load_index(local_index_path)
            _LOGGER.info(f"Using index {local_index_path} with {self.idx.element_count} points.")

            # load payloads:
            if isinstance(payloads, str):
                if payloads.endswith(".json"):
                    with open(payloads, "r") as f:
                        self.payloads = json.load(f)
                elif payloads.endswith(".pkl"):
                    self.payloads = pickle.load(open(payloads, "rb"))
                elif payloads.endswith(".yaml"):
                    with open(payloads, "r") as f:
                        self.payloads = yaml.load(f, Loader=yaml.SafeLoader)

                else:
                    raise ValueError(
                        f"payload should be either a json, pickle, or yaml file. you supplied: {payloads.split('.')[-1]}"
                    )
            else:
                self.payloads = payloads
            # self.payloads = {}
        # save the index to local file path
        else:
            _LOGGER.info(f"Index {local_index_path} does not exist, creating it.")
            self.idx.save_index(local_index_path)
            self.payloads = {}
        # self.payloads = payloads
        self.idx_path = local_index_path

    def load(
        self,
        vectors: np.ndarray,
        ids: Union[np.ndarray, None] = None,
        payloads: Union[List[Dict[str, str]], None] = None,
    ):
        """
        Upload embedding vectors into the hnsw index, and store their hnsw index id and payloads into metadata

        :param vectors: embedding vectors, a np.ndarray with shape of (n, <vector size>)
        :param ids: list of n point ids, or None to generate ids automatically
        :param payloads: optional list of n dictionaries that contain vector metadata
        :return:
        """

        # increase max_elements to contain new loadings
        current_max = self.idx.get_max_elements()

        if ids is None:
            new_max = current_max + vectors.shape[0]
            ids = np.arange(start=current_max, stop=new_max)
        else:
            new_max = ids.max() + 1

        # check if the number of embedding vectors and labels are same
        verify_load_inputs(vectors, ids, payloads)

        if payloads:
            for i in range(len(payloads)):
                self.payloads[ids[i]] = payloads[i]

        # update hnsw index and load embedding vectors
        self.idx.load_index(self.idx_path, max_elements=new_max)
        self.idx.add_items(vectors)

        # save hnsw index to local file
        self.idx.save_index(self.idx_path)

    def search(
        self,
        query: np.ndarray,
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> Union[
        List[Dict[str, Union[int, float, Dict[str, str], np.ndarray]]],
        List[List[Dict[str, Union[int, float, Dict[str, str], np.ndarray]]]],
    ]:
        """
        With query vector(s), get the limit nearest neighbors.

        :param query: the query vector, np.ndarray with shape of (1, dim) or (dim, )
        :param limit: number of nearest neighbors to search for query vector
        :param with_payload: whether payload is included in the result
        :param with_vectors: whether the stored vector is included in the result
        :param offset: the offset of the search results
        :return: if the shape of query vector is (<dim>, ), a list of limit dictionaries will be returned,
        the format of dictionary will be:
        {
            "id": <id>
            "distance": <distance>
            "payload": {
                <information of the vector>
            }
            "vector": [<the vector>]
        }
        if the shape of query vector is (n, <dim>), a 2d list will be returned,
        which is a list of n * list of limit dictionaries
        """
        ids, distances = self.idx.knn_query(query, k=limit + offset)
        # ids and distances are 2d array
        ids = ids.tolist()
        distances = distances.tolist()

        output_list = []
        for i in range(len(ids)):
            search_list = []
            result_id = ids[i]
            result_distances = distances[i]
            if with_vectors:
                result_vectors = self.idx.get_items(result_id, return_type="numpy")
            for j in range(limit):
                output_dict = {"id": result_id[j], "distance": result_distances[j]}
                if with_payload:
                    output_dict["payload"] = self.payloads[result_id[j]]
                if with_vectors:
                    output_dict["vector"] = result_vectors[j]
                search_list.append(output_dict)
            output_list.append(search_list)

        if len(output_list) == 1:
            return output_list[0]
        else:
            return output_list

    def __len__(self) -> int:
        return self.idx.element_count

    def retrieve_info(self, ids: Union[List[int], int], with_vectors: bool = False) -> Union[
        Dict[str, Union[int, List[float], Dict[str, str]]],
        List[Dict[str, Union[int, List[float], Dict[str, str]]]],
    ]:
        """
        With an id or a list of storage ids, return the information of these vectors
        :param ids: storage id, or a list of ids
        :param with_vectors: whether the stored vector is included in the result
        :return:
        """
        if not isinstance(ids, list):
            # retrieve() only takes iterable input
            ids = [ids]
        output_list = []
        for id_ in ids:
            output_dict = {"id": id_, "payload": self.payloads[id_]}
            output_list.append(output_dict)

        if with_vectors:
            vecs = self.idx.get_items(ids, return_type="numpy")
            for i in range(len(vecs)):
                output_list[i]["vector"] = vecs[i]

        # with just one id, only the dictionary instead of the list will be returned
        if len(output_list) == 1:
            return output_list[0]
        else:
            return output_list

    def __str__(self):
        return "HNSWBackend with {} items".format(len(self))

    def __repr__(self):
        return "HNSWBackend with {} items".format(len(self))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile
from multiprocessing import Pool

import numpy as np
import pandas as pd

from ..utils import natural_chr_sort
from .utils import check_if_uni_sorted, prep_data, process_line


def chrom_cmp(a, b):
    """Return smaller chromosome name"""
    c = natural_chr_sort(a, b)
    if c > 0:
        return b, True, False
    else:
        return a, False, True


def relationship_helper(region_a, region_b, only_in, overlap):
    """For two region calculate their overlap; for earlier region
    calculate how many base pair only in it
    :param [int, int] region_a: region that starts first
    :param [int, int] region_b: region that starts second
    :param int only_in: number of positions only in a so far
    :param int overlap: number of overlapping so far
    """
    if region_b[0] <= region_a[1]:
        only_in += region_b[0] - region_a[0]
        if region_b[1] <= region_a[1]:
            overlap += region_b[1] - region_b[0]
            start_a, start_b = region_b[1], region_b[1]
            inside_b, inside_a = False, True
            return only_in, inside_a, inside_b, overlap, start_a, start_b
        elif region_b[1] > region_a[1]:
            overlap += region_a[1] - region_b[0]
            start_a, start_b = region_a[1], region_a[1]
            inside_a, inside_b = False, True
            return only_in, inside_a, inside_b, overlap, start_a, start_b
    elif region_b[0] > region_a[1]:
        only_in += region_a[1] - region_a[0]
        inside_a, inside_b = False, True
        start_a, start_b = region_a[1], region_b[0]
        return only_in, inside_a, inside_b, overlap, start_a, start_b


def two_region_intersection_diff(
    region_d,
    region_q,
    only_in_d,
    only_in_q,
    inside_d,
    inside_q,
    overlap,
    start_d,
    start_q,
    waiting_d,
    waiting_q,
):
    """
    Check mutual position of two regions and calculate intersection and difference of two regions
    :param list region_d: region from universe
    :param list region_q: region from query
    :param int only_in_d: number of base pair only in universe
    :param int only_in_q: number of base pair only in query
    :param bool inside_d: whether there is still part of the region from universe to analyse
    :param bool inside_q: whether there is still part of the region from query to analyse
    :param int overlap: size of overlap
    :param int start_d: start position of currently analyzed universe region
    :param int start_q: start position of currently analyzed query region
    :param bool waiting_d: whether waiting for the query to finish chromosome
    :param bool waiting_q: whether waiting for the universe to finish chromosome
    """
    if waiting_q:
        only_in_d += region_d[1] - region_d[0]
        start_d = region_d[1]
        inside_d, inside_q = False, True
    elif waiting_d:
        only_in_q += region_q[1] - region_q[0]
        start_q = region_q[1]
        inside_q, inside_d = False, True
    else:
        if region_d[0] <= region_q[0]:
            res = relationship_helper(region_d, region_q, only_in_d, overlap)
            (only_in_d, inside_d, inside_q, overlap, start_d, start_q) = res
        if region_d[0] > region_q[0]:
            res = relationship_helper(region_q, region_d, only_in_q, overlap)
            (only_in_q, inside_q, inside_d, overlap, start_q, start_d) = res
    return only_in_d, only_in_q, inside_d, inside_q, overlap, start_d, start_q


def read_in_new_line(region, start, chrom, inside, waiting, lines, c_chrom, not_e):
    """
    Read in a new line from query or universe file
    """
    if not inside:
        if not waiting:
            line = lines.readline()
            if type(line) is bytes:
                line = line.decode("utf-8")
            line = line.strip("\n")
            if line != "":
                region, start, chrom = process_line(line)
                if chrom != c_chrom:
                    waiting = True
            else:
                not_e = False
                waiting = True
    return region, start, chrom, waiting, not_e


def calc_diff_intersection(db, folder, query):
    """
    Difference and overlap of two files on base pair level
    :param str db: path to universe file
    :param str folder: path to folder with query file
    :param str query: query file name
    :return str; int; int; int: file name; bp only in universe; bp only in query; overlap in bp
    """
    only_in_d, only_in_q, overlap = 0, 0, 0
    inside_d, inside_q = False, False  # inside a region
    not_end_d, not_end_q = True, True  # if there are regions to process
    waiting_d, waiting_q = (
        False,
        False,
    )  # if waiting for the other file to finish chrom
    lines_q = tempfile.NamedTemporaryFile()
    prep_data(folder, query, lines_q)
    if os.stat(lines_q.name).st_size == 0:
        print(f"Empty file {query}")
        lines_q.close()
        return [query, only_in_d, only_in_q, overlap]
    lines_db = open(db)
    new_d = lines_db.readline().strip("\n")
    pos_d, start_d, chrom_d = process_line(new_d)
    new_q = lines_q.readline().decode("utf-8").strip("\n")
    pos_q, start_q, chrom_q = process_line(new_q)
    if chrom_d == chrom_q:
        c_chrom = chrom_d
    else:
        c_chrom, waiting_d, waiting_q = chrom_cmp(chrom_d, chrom_q)
    while not_end_d or not_end_q:
        regions_stats = two_region_intersection_diff(
            [start_d, pos_d[1]],
            [start_q, pos_q[1]],
            only_in_d,
            only_in_q,
            inside_d,
            inside_q,
            overlap,
            start_d,
            start_q,
            waiting_d,
            waiting_q,
        )
        (
            only_in_d,
            only_in_q,
            inside_d,
            inside_q,
            overlap,
            start_d,
            start_q,
        ) = regions_stats
        new_d = read_in_new_line(
            pos_d,
            start_d,
            chrom_d,
            inside_d,
            waiting_d,
            lines_db,
            c_chrom,
            not_end_d,
        )
        (pos_d, start_d, chrom_d, waiting_d, not_end_d) = new_d
        new_q = read_in_new_line(
            pos_q,
            start_q,
            chrom_q,
            inside_q,
            waiting_q,
            lines_q,
            c_chrom,
            not_end_q,
        )
        (pos_q, start_q, chrom_q, waiting_q, not_end_q) = new_q
        if waiting_d or waiting_q:
            if not not_end_q:
                c_chrom = chrom_d
                waiting_d = False
            elif not not_end_d:
                c_chrom = chrom_q
                waiting_q = False
            elif chrom_d == chrom_q:
                c_chrom = chrom_d
                waiting_d, waiting_q = False, False
            elif waiting_d and waiting_q:
                c_chrom, waiting_d, waiting_q = chrom_cmp(chrom_d, chrom_q)

    only_in_d += pos_d[1] - start_d
    only_in_q += pos_q[1] - start_q
    lines_q.close()
    return [query, only_in_d, only_in_q, overlap]


def run_intersection(
    folder,
    file_list,
    universe,
    no_workers,
):
    """
    Calculate the base pair intersection of universe and group of files
    :param str folder: path to folder containing query files
    :param str file_list: path to file containing list of query files
    :param str universe: path to universe file
    :param int no_workers: number of parallel processes
    :param str save_to_file: whether to save median of calculated distances for each file
    :param str folder_out: output folder
    :param str pref: prefix used for saving
    :return float; float: mean of fractions of intersection of file and universe divided by universe size;
    mean of fractions of intersection of file and universe divided by file size
    """
    check_if_uni_sorted(universe)
    with open(file_list) as f:
        files = f.read().split("\n")[:-1]
    res = []
    if no_workers <= 1:
        for i in files:
            r = calc_diff_intersection(universe, folder, i)
            res.append(r)
    else:
        with Pool(no_workers) as p:
            args = [(universe, folder, f) for f in files]
            res = p.starmap(calc_diff_intersection, args)
    return pd.DataFrame(res)



import gzip
import os
import shutil
from contextlib import suppress
from logging import getLogger
from typing import Dict, List, NoReturn, Union

import boto3
import requests
import s3fs
import zarr
from botocore.exceptions import ClientError
from pybiocfilecache import BiocFileCache
from pybiocfilecache.exceptions import RnameExistsError
from ubiquerg import is_url
from zarr import Array
from zarr.errors import PathNotFoundError

from ..exceptions import TokenizedFileNotFoundError, TokenizedFileNotFoundInCacheError
from ..io.io import BedSet, RegionSet
from ..io.utils import is_gzipped
from .const import (
    BED_TOKENS_PATTERN,
    BEDFILE_URL_PATTERN,
    BEDSET_URL_PATTERN,
    DEFAULT_BEDBASE_API,
    DEFAULT_BEDFILE_EXT,
    DEFAULT_BEDFILE_SUBFOLDER,
    DEFAULT_BEDSET_EXT,
    DEFAULT_BEDSET_SUBFOLDER,
    DEFAULT_BUCKET_FOLDER,
    DEFAULT_BUCKET_NAME,
    DEFAULT_CACHE_FOLDER,
    DEFAULT_ZARR_FOLDER,
    MODULE_NAME,
)
from .utils import BedCacheManager, get_abs_path

_LOGGER = getLogger(MODULE_NAME)


class BBClient(BedCacheManager):
    def __init__(
        self,
        cache_folder: Union[str, os.PathLike] = DEFAULT_CACHE_FOLDER,
        bedbase_api: str = DEFAULT_BEDBASE_API,
    ):
        """
        BBClient to deal with download files from bedbase and caching them.

        :param cache_folder: path to local folder as cache of files from bedbase,
        if not given it will be the environment variable `BBCLIENT_CACHE`
        :param bedbase_api: url to bedbase
        """
        cache_folder = get_abs_path(cache_folder)
        super().__init__(cache_folder)

        self._bedfile_cache = BiocFileCache(os.path.join(cache_folder, DEFAULT_BEDFILE_SUBFOLDER))
        self._bedset_cache = BiocFileCache(os.path.join(cache_folder, DEFAULT_BEDSET_SUBFOLDER))

        self.zarr_cache = zarr.group(
            store=os.path.join(cache_folder, DEFAULT_ZARR_FOLDER), overwrite=False
        )
        self.bedbase_api = bedbase_api

    def load_bedset(self, bedset_id: str) -> BedSet:
        """
        Load a BEDset from cache, or download and add it to the cache with its BED files

        :param bedset_id: unique identifier of a BED set
        :return: the BedSet object
        """

        file_path = self._bedset_path(bedset_id)

        if os.path.exists(file_path):
            _LOGGER.info(f"BED set {bedset_id} already exists in cache.")
            with open(file_path, "r") as file:
                extracted_data = file.read().splitlines()
        else:
            extracted_data = self._download_bedset_data(bedset_id)
            # write the identifiers of BED files in the BedSet to a local .txt file
            with open(file_path, "w") as file:
                for value in extracted_data:
                    file.write(value + "\n")
            _LOGGER.info(f"BED set {bedset_id} downloaded and cached successfully.")

        # return the BedSet
        return BedSet(
            [self.load_bed(bedfile_id) for bedfile_id in extracted_data],
            identifier=bedset_id,
        )

    def _download_bedset_data(self, bedset_id: str) -> List[str]:
        """
        Download BED set from BEDbase API and return the list of identifiers of BED files in the set

        :param bedset_id: unique identifier of a BED set
        :return: the list of identifiers of BED files in the set
        """
        bedset_url = BEDSET_URL_PATTERN.format(bedbase_api=self.bedbase_api, bedset_id=bedset_id)
        response = requests.get(bedset_url)
        data = response.json()["results"]
        extracted_data = [entry.get("id") for entry in data]

        return extracted_data

    def load_bed(self, bed_id: str) -> RegionSet:
        """
        Loads a BED file from cache, or downloads and caches it if it doesn't exist

        :param bed_id: unique identifier of a BED file
        :return: the RegionSet object
        """
        file_path = self._bedfile_path(bed_id)

        if os.path.exists(file_path):
            _LOGGER.info(f"BED file {bed_id} already exists in cache.")
        # if not in the cache, download from BEDbase and write to file in cache
        else:
            bed_data = self._download_bed_file_from_bb(bed_id)
            with open(file_path, "wb") as f:
                f.write(bed_data)

            self._bedfile_cache.add(bed_id, fpath=file_path, action="asis")

            _LOGGER.info(f"BED file {bed_id} was downloaded and cached successfully")

        return RegionSet(regions=file_path)

    def add_bedset_to_cache(self, bedset: BedSet) -> str:
        """
        Add a BED set to the cache

        :param bedset: the BED set to be added, a BedSet class
        :return: the identifier if the BedSet object
        """

        bedset_id = bedset.compute_bedset_identifier()
        file_path = self._bedset_path(bedset_id)
        if os.path.exists(file_path):
            _LOGGER.info(f"{file_path} already exists in cache.")
        else:
            with open(file_path, "w") as file:
                for bedfile in bedset:
                    bedfile_id = self.add_bed_to_cache(bedfile)
                    file.write(bedfile_id + "\n")
        self._bedset_cache.add(bedset_id, fpath=file_path, action="asis")
        return bedset_id

    def add_bed_to_cache(self, bedfile: Union[RegionSet, str], force: bool = False) -> str:
        """
        Add a BED file to the cache

        :param bedfile: a RegionSet object or a path or url to the BED file
        :param force: whether to overwrite the existing file in cache
        :return: the RegionSet identifier
        """

        if isinstance(bedfile, str):
            bedfile = RegionSet(bedfile)
        elif not isinstance(bedfile, RegionSet):
            raise TypeError(
                f"Input must be a RegionSet or a path to a BED file, not {type(bedfile)}"
            )

        bedfile_id = bedfile.compute_bed_identifier()
        file_path = self._bedfile_path(bedfile_id)
        if os.path.exists(file_path) and not force:
            _LOGGER.info(f"{file_path} already exists in cache.")
        else:
            if bedfile.path is None or is_url(bedfile.path):
                bedfile.to_pandas().to_csv(
                    file_path, index=False, compression="gzip", header=False, sep="\t"
                )
            else:
                # copy the BED file out of cache
                if is_gzipped(bedfile.path):
                    shutil.copyfile(bedfile.path, file_path)
                else:
                    # https://docs.python.org/3/library/gzip.html
                    with open(bedfile.path, "rb") as f_in:
                        with gzip.open(file_path, "wb") as f_out:
                            shutil.copyfileobj(f_in, f_out)
            with suppress(RnameExistsError):
                self._bedfile_cache.add(bedfile_id, fpath=file_path, action="asis")
        return bedfile_id

    def add_bed_tokens_to_cache(self, bed_id: str, universe_id: str) -> None:
        """
        Add a tokenized BED file to the cache

        :param bed_id: the identifier of the BED file
        :param universe_id: the identifier of the universe

        :return: the identifier of the tokenized BED file
        """

        tokens_info_url = BED_TOKENS_PATTERN.format(
            bedbase_api=DEFAULT_BEDBASE_API, bed_id=bed_id, universe_id=universe_id
        )
        response = requests.get(tokens_info_url)
        if response.status_code == 404:
            raise TokenizedFileNotFoundError(
                f"Tokenized BED file {bed_id} for {universe_id} does not exist in bedbase."
                f"Please make sure the tokenized BED file is available in bedbase."
            )

        tokens_info = response.json()
        file_path = tokens_info["file_path"]

        s3fc_obj = s3fs.S3FileSystem(endpoint_url=tokens_info["endpoint_url"])
        zarr_store = s3fs.S3Map(root=file_path, s3=s3fc_obj, check=False, create=False)
        cache_obj = zarr.LRUStoreCache(zarr_store, max_size=2**28)

        try:
            tokenized_bed = zarr.open(cache_obj, mode="r")
        except PathNotFoundError:
            raise TokenizedFileNotFoundError(
                f"Tokenized BED file {bed_id} for {universe_id} does not exist in bedbase."
                f"Please make sure the tokenized BED file is available in bedbase."
            )

        self.cache_tokens(bed_id, universe_id, tokenized_bed)

    def load_bed_tokens(self, bed_id: str, universe_id: str) -> Array:
        """
        Load a tokenized BED file from cache, or download and cache it if it doesn't exist

        :param bed_id: the identifier of the BED file
        :param universe_id: the identifier of the universe

        :return: the zarr array of tokens
        """
        try:
            zarr_array = self.zarr_cache[universe_id][bed_id]
        except KeyError:
            try:
                self.add_bed_tokens_to_cache(bed_id, universe_id)
            except TokenizedFileNotFoundError:
                raise TokenizedFileNotFoundInCacheError(
                    f"Tokenized BED file {bed_id} for {universe_id} does not exist in cache."
                    "And it is not available in bedbase."
                )
            zarr_array = self.zarr_cache[universe_id][bed_id]

        return zarr_array

    def remove_tokens(self, bed_id: str, universe_id: str) -> None:
        """
        Remove all tokenized BED files from cache
        """
        try:
            del self.zarr_cache[universe_id][bed_id]
        except KeyError:
            raise TokenizedFileNotFoundInCacheError(
                f"Tokenized BED file {bed_id} for {universe_id} does not exist in cache."
            )

    def cache_tokens(self, bed_id: str, universe_id: str, tokens: Union[list, Array]) -> None:
        """
        Cache tokenized BED file

        :param bed_id: the identifier of the BED file
        :param universe_id: the identifier of the universe
        :param tokens: the list of tokens

        :return: None
        """

        univers_group = self.zarr_cache.require_group(universe_id)
        univers_group.create_dataset(bed_id, data=tokens, overwrite=True)

        _LOGGER.info(
            f"Tokenized BED file {bed_id} tokenized using {universe_id} was cached successfully"
        )

    def add_bed_to_s3(
        self,
        identifier: str,
        bucket: str = DEFAULT_BUCKET_NAME,
        endpoint_url: str = None,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        s3_path: str = DEFAULT_BUCKET_FOLDER,
    ) -> str:
        """
        Add a cached BED file to S3

        :param identifier: the unique identifier of the BED file
        :param bucket: the name of the bucket
        :param endpoint_url: the URL of the S3 endpoint [Default: set up by the environment vars]
        :param aws_access_key_id: the access key of the AWS account [Default: set up by the environment vars]
        :param aws_secret_access_key: the secret access key of the AWS account [Default: set up by the environment vars]
        :param s3_path: the path on S3

        :return: full path on S3
        """
        s3_client = boto3.client(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        local_file_path = self.seek(identifier)
        bed_file_name = os.path.basename(local_file_path)
        s3_bed_path = os.path.join(identifier[0], identifier[1], bed_file_name)
        if s3_path:
            s3_bed_path = os.path.join(s3_path, s3_bed_path)

        s3_client.upload_file(local_file_path, bucket, s3_bed_path)
        _LOGGER.info(f"Project was uploaded successfully to s3://{bucket}/{s3_bed_path}")
        return s3_bed_path

    def get_bed_from_s3(
        self,
        identifier: str,
        bucket: str = DEFAULT_BUCKET_NAME,
        endpoint_url: str = None,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        s3_path: str = DEFAULT_BUCKET_FOLDER,
    ) -> str:
        """
        Get a cached BED file from S3 and cache it locally

        :param identifier: the unique identifier of the BED file
        :param bucket: the name of the bucket
        :param endpoint_url: the URL of the S3 endpoint [Default: set up by the environment vars]
        :param aws_access_key_id: the access key of the AWS account [Default: set up by the environment vars]
        :param aws_secret_access_key: the secret access key of the AWS account [Default: set up by the environment vars]
        :param s3_path: the path on S3

        :return: bed file id
        :raise FileNotFoundError: if the identifier does not exist in cache
        """
        s3_bed_path = os.path.join(
            identifier[0], identifier[1], f"{identifier}{DEFAULT_BEDFILE_EXT}"
        )
        if s3_path:
            s3_bed_path = os.path.join(s3_path, s3_bed_path)

        s3_client = boto3.client(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        try:
            s3_client.download_file(
                bucket, s3_bed_path, self._bedfile_path(identifier, create=True)
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                raise FileNotFoundError(f"{identifier} does not exist in S3.")
            else:
                raise e

        return identifier

    def seek(self, identifier: str) -> str:
        """
        Get local path to BED file or BED set with specific identifier

        :param identifier: the unique identifier
        :return: the local path of the file
        :raise FileNotFoundError: if the identifier does not exist in cache
        """

        # bedfile
        file_path = self._bedset_path(identifier, False)
        if os.path.exists(file_path):
            return file_path
        else:
            # bedset
            file_path = self._bedfile_path(identifier, False)
            if os.path.exists(file_path):
                return file_path
            else:
                raise FileNotFoundError(f"{identifier} does not exist in cache.")

    def remove_bedset_from_cache(self, bedset_id: str, remove_bed_files: bool = False) -> NoReturn:
        """
        Remove a BED set from cache

        :param bedset_id: the identifier of BED set
        :param remove_bed_files: whether also remove BED files in the BED set
        :raise FileNotFoundError: if the BED set does not exist in cache
        """

        file_path = self.seek(bedset_id)
        if remove_bed_files:
            with open(file_path, "r") as file:
                extracted_data = file.readlines()
            for bedfile_id in extracted_data:
                self.remove_bedfile_from_cache(bedfile_id)

        self._bedset_cache.remove(bedset_id)
        # commented due to bioc file cache removal:
        # self._remove(file_path)

    def list_beds(self) -> Dict[str, str]:
        """
        List all BED files in cache

        :return: the list of identifiers of BED files
        """

        resources = self._bedfile_cache.list_resources()

        results = {}
        for resource in resources:
            results[resource.rname] = resource.fpath

        results = dict(sorted(results.items()))
        return results

    def list_bedsets(self) -> Dict[str, str]:
        """
        List all BED sets in cache

        :return: the list of identifiers of BED sets
        """

        resources = self._bedset_cache.list_resources()

        results = {}
        for resource in resources:
            results[resource.rname] = resource.fpath

        results = dict(sorted(results.items()))
        return results

    def _download_bed_file_from_bb(self, bedfile: str) -> bytes:
        """
        Download BED file from BEDbase API and return the file content as bytes

        :param bedfile: unique identifier of a BED file
        :return: the file content as bytes
        """

        bed_url = BEDFILE_URL_PATTERN.format(bedbase_api=self.bedbase_api, bed_id=bedfile)
        response = requests.get(bed_url)
        response.raise_for_status()
        return response.content

    def _bedset_path(self, bedset_id: str, create: bool = True) -> str:
        """
        Get the path of a BED set's .txt file with given identifier

        :param bedset_id: the identifier of BED set
        :param create: whether the cache path needs creating
        :return: the path to the .txt file
        """

        subfolder_name = DEFAULT_BEDSET_SUBFOLDER
        file_extension = DEFAULT_BEDSET_EXT

        return self._cache_path(bedset_id, subfolder_name, file_extension, create)

    def _bedfile_path(self, bedfile_id: str, create: bool = True) -> str:
        """
        Get the path of a BED file's .bed.gz file with given identifier

        :param bedfile_id: the identifier of BED file
        :param create: whether the cache path needs creating
        :return: the path to the .bed.gz file
        """

        subfolder_name = DEFAULT_BEDFILE_SUBFOLDER
        file_extension = DEFAULT_BEDFILE_EXT

        return self._cache_path(bedfile_id, subfolder_name, file_extension, create)

    def _cache_path(
        self,
        identifier: str,
        subfolder_name: str,
        file_extension: str,
        create: bool = True,
    ) -> str:
        """
        Get the path of a file in cache folder

        :param identifier: the identifier of BED set or BED file
        :param subfolder_name: "bedsets" or "bedfiles"
        :param file_extension: ".txt" or ".bed.gz"
        :param create: whether the cache path needs creating
        :return: the path to the file
        """

        filename = f"{identifier}{file_extension}"
        folder_name = os.path.join(self.cache_folder, subfolder_name, identifier[0], identifier[1])
        if create:
            self.create_cache_folder(folder_name)
        return os.path.join(folder_name, filename)

    def remove_bedfile_from_cache(self, bedfile_id: str) -> NoReturn:
        """
        Remove a BED file from cache

        :param bedfile_id: the identifier of BED file
        :raise FileNotFoundError: if the BED set does not exist in cache
        """

        # commented due to bioc chacing removal method
        # file_path = self.seek(bedfile_id)
        # self._remove(file_path)
        self._bedfile_cache.remove(bedfile_id)

    @staticmethod
    def _remove(file_path: str) -> None:
        """
        Remove a file within the cache with given path, and remove empty subfolders after removal
        Structure of folders in cache:
        cache_folder
            bedfiles
                a/b/ab1234xyz.bed.gz
            bedsets
                c/d/cd123hij.txt

        :param file_path: the path to the file
        :return: None
        """
        # the subfolder that matches the second digit of the identifier
        sub_folder_2 = os.path.split(file_path)[0]
        # the subfolder that matches the first digit of the identifier
        sub_folder_1 = os.path.split(sub_folder_2)[0]

        os.remove(file_path)

        # if the subfolders are empty after removal, remove the folders too
        if len(os.listdir(sub_folder_2)) == 0:
            os.rmdir(sub_folder_2)
            if len(os.listdir(sub_folder_1)) == 0:
                os.rmdir(sub_folder_1)

        return None



import pytest
import scanpy as sc

from geniml.io.io import RegionSet
from geniml.tokenization.main import AnnDataTokenizer, TreeTokenizer


@pytest.fixture
def pbmc_data():
    return sc.read_h5ad("tests/data/pbmc_hg38.h5ad")


@pytest.fixture
def pbmc_data_backed():
    return sc.read_h5ad("tests/data/pbmc_hg38.h5ad", backed="r")


@pytest.fixture
def universe_bed_file():
    return "tests/data/universe.uniq.bed"


def test_create_universe(universe_bed_file: str):
    u = RegionSet(universe_bed_file)
    assert u is not None
    assert len(u) == 2_378


def test_make_tree_tokenizer(universe_bed_file: str):
    t = TreeTokenizer(universe_bed_file)
    assert t is not None


def test_tree_tokenize_region_set(universe_bed_file: str):
    t = TreeTokenizer(universe_bed_file)
    assert t is not None

    # tokenize a bed file
    bed_file = "tests/data/to_tokenize.bed"

    # read in the bed file to test
    rs = RegionSet(bed_file)

    # tokenize
    tokens = t.tokenize(rs)
    region_tokens = tokens

    # filter out UNK tokens
    assert len(region_tokens) == 14  # one of the regions gets split into two tokens
    region_tokens = [t for t in region_tokens if t.chr != "chrUNK"]

    # count tokens
    assert len([t for t in region_tokens if t is not None]) == 3


def test_yield_tokens(universe_bed_file: str):
    t = TreeTokenizer(universe_bed_file)

    # tokenize a bed file
    bed_file = "tests/data/to_tokenize.bed"

    # read in the bed file to test
    rs = RegionSet(bed_file)

    # tokenize
    tokens = t(rs)
    region_tokens = tokens.to_regions()

    # filter out UNK tokens
    assert len(region_tokens) == 14  # one of the regions gets split into two tokens
    region_tokens = [t for t in region_tokens if t.chr != "chrUNK"]

    # count tokens
    assert len([t for t in region_tokens if t is not None]) == 3

    # yield tokens
    for token in tokens:
        assert token is not None
        assert isinstance(token.id, int)


def test_tokenize_anndata(universe_bed_file: str):
    t = AnnDataTokenizer(universe_bed_file)

    # tokenize anndata
    adata = sc.read_h5ad("tests/data/pbmc_hg38.h5ad")

    # tokenize
    tokens = t.tokenize(adata)

    # count tokens
    assert len(tokens) == 20
